import sys, cmd
from rply.errors import LexingError

from 木兰.分析器.词法分析器 import *
from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.环境 import 创建全局变量
from 木兰.功用.反馈信息 import 反馈信息
from 木兰.分析器.错误 import 语法错误, 词法错误

# TODO: 更多测试
def 括号已配对(源码):
    关键词 = {
        名词_操作符, 名词_函数, 名词_类型, #'OPERATOR', 'FUNC', 'ATTR', 'TYPE',
        连词_对, 动词_循环, 连词_每当,
        连词_如果, 连词_否则如果, 连词_否则,
        #'TRY', 'CATCH', 'FINALLY'
    }
    未配对之和 = 0
    未闭合的个数 = 0
    if len(源码) > 1:
        if 源码[-2] == '\\':
            return False
        else:
            各词 = 分词器.lex(源码)
            未闭合的 = []
            未配对 = [0, 0, 0]
            末两词 = 2 * ['']
            for 词 in 各词:
                c = 词.gettokentype()
                末两词[0], 末两词[1] = 末两词[1], c
                if c in 关键词:
                    未闭合的.append(c)
                if c == 前大括号:
                    未配对[0] += 1
                elif c == 后大括号:
                    未配对[0] -= 1
                    if len(未闭合的):
                        未闭合的.pop(-1)
                elif c == 前小括号:
                    未配对[1] += 1
                elif c == 后小括号:
                    未配对[1] -= 1
                elif c == 前中括号:
                    未配对[2] += 1
                elif c == 后中括号:
                    未配对[2] -= 1
            未配对之和 = sum(未配对)
            未闭合的个数 = len(未闭合的)
            if 未闭合的个数 > 0:
                if 未配对之和 == 0:
                    if 末两词[1] == 换行:
                        if 末两词[0] == 换行 or 末两词[0] == 分号:
                            return True
    return 未闭合的个数 == 0 and 未配对之和 == 0


def input_swallowing_interrupt(_input):

    def _input_swallowing_interrupt(*args):
        try:
            return _input(*args)
        except KeyboardInterrupt:
            # TODO: 何用？
            print('^C')
            return '\n'

    return _input_swallowing_interrupt


class 交互(cmd.Cmd):
    # 本地变量暂时无用
    def __init__(self, 提示符1='> ', 提示符2='>> ', 全局变量=None):
        super().__init__()
        self.提示符1 = 提示符1
        self.提示符2 = 提示符2
        self.全局变量 = 全局变量
        self.分析器 = 语法分析器(分词器)
        self.prompt = 提示符1
        self.声明 = ''

    def do_EOF(self, arg):
        self.default('再会()')

    def onecmd(self, 行):
        if 行 == 'EOF':
            return self.do_EOF(行)
        self.default(行)
        self.prompt = self.提示符1 if len(self.声明) == 0 else self.提示符2

    def default(self, 行):
        if 行 is not None:
            self.声明 += '%s\n' % 行
            try:
                if not self.括号已配对():
                    return
            except LexingError as e:
                错误 = 词法错误(异常=e,
                    文件名='【标准输入】',
                    源码=self.声明)
                sys.stderr.write('%s\n' % 反馈信息(错误))
                self.声明 = ''
            try:
                try:
                    节点 = self.分析器.分析('___=(%s);__print__(___)' % self.声明, '【标准输入】')
                except Exception:
                    节点 = self.分析器.分析(self.声明, '【标准输入】')

                可执行码 = compile(节点, '【标准输入】', 'exec')
                exec(可执行码, self.全局变量)
            except SystemExit:
                sys.exit()
            except BaseException as e:
                try:
                    sys.stderr.write('%s\n' % 反馈信息(e))
                finally:
                    e = None
                    del e

            finally:
                self.声明 = ''

    def 括号已配对(self):
        return 括号已配对(self.声明)

    def cmdloop(self, *args, **kwargs):
        原始输入函数 = cmd.__builtins__['input']
        cmd.__builtins__['input'] = input_swallowing_interrupt(原始输入函数)
        try:
            (super().cmdloop)(*args, **kwargs)
        finally:
            cmd.__builtins__['input'] = 原始输入函数


def 开始交互(提示符1='> ', 提示符2='>> ', 全局变量=None):
    """
    简易的木兰交互环境
    """
    介绍 = [
     '\t详情: 列出内置功能',
     '\t再会: 结束对话',
     '\t你好: 显示这段']
    if not 全局变量:
        # TODO: 何时会用
        全局变量 = 创建全局变量(文件名='<STDIN>')
    全局变量['详情'] = lambda : print('\n'.join([' %s (%s)' % (k, v.__class__.__name__) for k, v in 全局变量.items() if k != '__builtins__' if k != '___']))
    全局变量['你好'] = lambda *args: print('\n'.join(介绍)) if not args else print()
    交互(提示符1, 提示符2, 全局变量).cmdloop("木兰向您问好\n更多信息请说'你好'")
    sys.exit(0)


import ast

class 语法树相关:

    def 格式化节点(节点, 层次):
        缩进 = "  "
        输出 = ""
        if isinstance(节点, list):
            输出 += "["
            for 子节点 in 节点:
                输出 += 语法树相关.格式化节点(子节点, 层次 + 1)
            输出 += "]"
        elif isinstance(节点, int):
            输出 += str(节点)
        elif isinstance(节点, str):
            输出 += "'" + 节点 + "'"
        else:
            if 节点 == None:
              return 输出 + "None"
            输出 += type(节点).__name__ + "("
            属性个数 = 0
            for 属性 in ast.iter_fields(节点):
                属性个数 += 1
                输出 += "\n" + 缩进 * 层次 + 属性[0] + "="
                输出 += 语法树相关.格式化节点(属性[1], 层次 + 1)
            if isinstance(节点, ast.stmt) or isinstance(节点, ast.expr):
                输出 += f"\n{缩进 * 层次}lineno={节点.lineno}"
                输出 += f"\n{缩进 * 层次}col_offset={节点.col_offset}"
            if 属性个数 == 0:
                return 输出 + ")"
            return 输出 + "\n" + 缩进 * (层次 - 1) + ")"
        return 输出

'''
TODO:
如果class的decorator_list为空(无NameFixPass), 则报错:
Traceback (most recent call last):
  File "./中.py", line 26, in <module>
    print(语法树相关.格式化节点(节点, 1))
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 23, in 格式化节点
    输出 += 语法树相关.格式化节点(属性[1], 层次 + 1)
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 11, in 格式化节点
    输出 += 语法树相关.格式化节点(子节点, 层次 + 1)
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 23, in 格式化节点
    输出 += 语法树相关.格式化节点(属性[1], 层次 + 1)
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 11, in 格式化节点
    输出 += 语法树相关.格式化节点(子节点, 层次 + 1)
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 23, in 格式化节点
    输出 += 语法树相关.格式化节点(属性[1], 层次 + 1)
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 23, in 格式化节点
    输出 += 语法树相关.格式化节点(属性[1], 层次 + 1)
  File "/Users/xuanwu/work/木兰/prototype/功用.py", line 20, in 格式化节点
    for 属性 in ast.iter_fields(节点):
  File "/opt/miniconda3/lib/python3.7/ast.py", line 177, in iter_fields
    for field in node._fields:
AttributeError: 'NoneType' object has no attribute '_fields'
'''


class 语法错误(ValueError):

    def __init__(self, 信息, 文件名, 行号, 列号, 源码=None):
        self.信息 = 信息
        self.文件名 = 文件名
        self.行号 = 行号 if 行号 > 0 else 1
        self.列号 = 列号 if 列号 > 0 else 1
        self.源码 = 源码

    # TODO: 应使错误更易定位
    def __str__(self):
        反馈信息 = '文件 "%s", 第%d行, 第%d列, %s' % (
                self.文件名, self.行号, self.列号, self.信息)
        if self.源码:
            行 = self.源码[(self.行号 - 1)]
            出错位置 = self.列号 - 1
        return '%s\n%s' % (反馈信息, 行[:出错位置] + '✋' + 行[出错位置:])

class 词法错误(语法错误):

    def __init__(self, 异常, 文件名, 源码=None):
        self.信息='分词时没认出这个词 "%s"' % 源码[异常.getsourcepos().idx]
        self.文件名=文件名
        self.行号=异常.getsourcepos().lineno
        self.列号=异常.getsourcepos().colno
        self.源码=源码.split("\n")


from sys import platform

为win系统 = platform == 'win32'

期望值 = {
    "运算/加.ul": b'5',
    "运算/减.ul": b"1",
    "运算/乘.ul": b"6181",
    "运算/除整.ul": b"2",
    "运算/除留整.ul": b"10",
    "运算/四则运算.ul": b"4",
    "运算/加小数.ul": b"5.0",
    "运算/除小数.ul": b"2.0",
    "运算/比较.ul": b"2222222222222",
    "运算/赋值.ul": b"2",
    "运算/赋值两次.ul": b"6",
    "运算/赋值增量.ul": b"21",
    "运算/一元操作.ul": b"2-1-2",
    "运算/赋值多项.ul": b"2112123",
    "运算/其他.ul": b"82",

    # TODO: 加上 `(2<1)!=nil and`
    "运算/空.ul": b"truetrue",

    "流程控制/条件.ul": b"31",
    "流程控制/条件否则如果.ul": b"2",
    "流程控制/条件否则.ul": b"4",
    "流程控制/条件三段.ul": b"5",
    "流程控制/条件倒置.ul": b"65",
    "流程控制/条件否则多行.ul": b"34",
    "流程控制/每当.ul": b"6",
    "流程控制/循环控制.ul": b"2301201245",
    "流程控制/三元表达式.ul": b"213",
    "流程控制/循环for.ul": b"0120120212301251a2b",
    "流程控制/空块.ul": b"123",
    "流程控制/循环loop.ul": b"123",
    "流程控制/try.ul": b"030",
    "流程控制/catch.ul": b"vi",
    "流程控制/catch名称.ul": b"list index out of range",

    "函数/无参数.ul": b"223", # 最后无4是因为返回了函数, 而非它的返回值
    "函数/单参数.ul": b"2",
    "函数/多参数.ul": b"123",
    "函数/多层调用.ul": b"3",
    "函数/二阶函数.ul": b"11",
    "函数/返回空.ul": b"2",
    "函数/返回单值.ul": b"2",
    "函数/全局.ul": b"42",
    "函数/无返回.ul": b"2nil",
    "函数/过滤.ul": b"[10]",
    "函数/map.ul": b"[1, 4, 9]",
    "函数/返回多值.ul": b"12", # TODO: 原本元组输出为 1, 2， 而非(1，2)
    "函数/匿名函数.ul": b"[1][1][3][3][1, 4]12true",
    "函数/API/内置.ul": b"truefalsea[1]200 a1 b2 ctruefalse32",
    "函数/API/文件.ul": b"hi",
    "函数/API/self.ul": b"true",
    "函数/API/file.ul": b"true",
    "函数/形参默认值.ul": b"2",
    "函数/指名参数.ul": b"466",
    "函数/指定返回类型.ul": b"Mulan1",
    "函数/指定形参类型.ul": b"hi Mulan1",
    # "函数/变长参数.ul": b"1 2 3",

    "特殊字符/多行.ul": b"23", # TODO: 如果末尾加空行, 报错 rply.errors.ParsingError: (None, None)
    "特殊字符/块.ul": b"2",
    "特殊字符/块多行.ul": b"3",
    "特殊字符/空格.ul": b"2",
    "特殊字符/制表符.ul": b"2",
    "特殊字符/缩进.ul": b"2",
    "特殊字符/注释块.ul": b"23",
    "特殊字符/中文标识符.ul": b"2020",
    "特殊字符/空行.ul": b"1",
    "特殊字符/小括号.ul": b"4",
    "特殊字符/分号.ul": b"2",
    "特殊字符/美元.ul": b"566787",

    "引用/引用本地py.ul": b"2",
    "引用/引用本地包内py.ul": b"2",
    "引用/引用本地多py.ul": b"23",
    "引用/引用本地py全部内容.ul": b"32",

    # TODO： 深究 python 中 'from . import *'的含义。参考：
    # https://stackoverflow.com/questions/57774193/what-does-from-dot-import-asterisk-do-in-python-3
    # https://blog.csdn.net/nigelyq/article/details/78930330
    # "引用/引用本地py点.ul": b"2",

    # TODO: 引用 python 标准库, 第三方库
    "引用/引用标准py_math.ul": b"5",
    "引用/引用标准py_random.ul": b"true",

    "引用/引用本地py某内容.ul": b"2",
    "引用/引用木兰多个.ul": b"21",
    "引用/引用木兰全部内容.ul": b"2",
    "引用/引用本地包内木兰.ul": b"2",
    "引用/引用本地包内木兰某内容.ul": b"2",

    "引用/枚举/b.ul": b"T.Xfalsetrue",

    "类型/定义.ul": b"true",
    "类型/定义静态方法.ul": b"11",
    "类型/定义方法.ul": b"1",
    "类型/定义方法self.ul": b"11",
    "类型/定义构造方法.ul": b"1",
    "类型/构造方法原始.ul": b"Mulan",
    "类型/个体属性.ul": b"111",
    "类型/类型属性.ul": b"2[2]",
    "类型/应变属性.ul": b"343",
    "类型/继承.ul": b"1",
    "类型/继承于调用.ul": b"1",
    "类型/操作符/定义操作符.ul": b"31true",
    "类型/超类.ul": b"animal",
    "类型/返回值属性.ul": b"ok",

    "字符串/双引号.ul": b'ok\n\t\\"\\natruetrue' if not 为win系统 else b'ok\r\n\t\\"\\natruetrue',
    "字符串/单引号.ul": b"ok\n\t\\'\\n" if not 为win系统 else b"ok\r\n\t\\'\\n",
    "字符串/相关方法.ul": b"1hi[fish]",
    "字符串/插值.ul": b"b3b3o44ta3a44t`a3a",

    "数据结构/范围.ul": b"range(0, 3)range(0, 3)range(0, 2)range(-1, 5, 2)range(4, -2, -2)02",
    "数据结构/列表.ul": b"[][2][2, 4, 6][[1], [2]]",
    "数据结构/列表取值.ul": b"245[1, 4][3, 5][1, 4][1, 4, 3, 5]",
    "数据结构/列表组合.ul": b"[a, b, b]",
    "数据结构/字典.ul": b"0acfalsefalse1a1b21",
    "数据结构/集合.ul": b"{1, 2, 3}",
    "数据结构/多项.ul": b"1, 2, 3",
    "数据结构/枚举.ul": b"12",
    "数据结构/标识符.ul": b"36",

    "算法/排序/冒泡.ul": b"[1, 2, 4, 5, 8]",
    "算法/排序/插入.ul": b"[1, 2, 4, 5, 8]",
    "算法/排序/快速.ul": b"[1, 2, 4, 5, 8]",

    "综合.ul": b"10",
    #"错误处理/死递归.ul": b"test",
}



#!/usr/bin/env python3

import shutil
import subprocess

from 期望值表 import *

# 原始可执行文件在：https://gitee.com/MulanRevive/bounty/tree/master/%E5%8E%9F%E5%A7%8B%E8%B5%84%E6%96%99/%E5%8F%AF%E6%89%A7%E8%A1%8C%E6%96%87%E4%BB%B6/ulang-0.2.2.exe
# 原始版本不支持中文标识符，且仅支持 gbk 编码的源文件。因此英文版测试文件仅用英文字符串和标识符。

路径 = '测试/'
# 不确定为何输出是bytes：https://stackoverflow.com/questions/6269765/what-does-the-b-character-do-in-front-of-a-string-literal

# 多进程参考：https://shuzhanfan.github.io/2017/12/parallel-processing-python-subprocess/
进程表 = {}
原始可执行文件 = "ulang-0.2.2.exe"
木兰重现 = "木兰"

可执行文件 = 原始可执行文件 if 为win系统 else 木兰重现

if shutil.which(可执行文件) is None:
    print("运行测试需安装'" + 可执行文件 + "'")
    exit()

英文版 = set(["运算/乘.ul",
    "函数/过滤.ul", "函数/map.ul", "函数/返回多值.ul", "函数/匿名函数.ul", "函数/API/文件.ul",
    "特殊字符/注释块.ul",
    "类型/继承.ul", "类型/个体属性.ul", "类型/定义方法self.ul",
    "字符串/双引号.ul", "字符串/相关方法.ul",
    "数据结构/列表取值.ul", "数据结构/枚举.ul",
    "引用/引用木兰多个.ul", "引用/引用本地包内木兰.ul", "引用/引用本地包内木兰某内容.ul",
    "算法/排序/冒泡.ul", "算法/排序/插入.ul", "算法/排序/快速.ul",
    "综合.ul"])
# 参考：https://stackoverflow.com/questions/748028/how-to-get-output-of-exe-in-python-script
for 文件 in 期望值:
    源码路径 = 路径 + 文件
    if 为win系统:
        if 文件 in 英文版:
            源码路径 = 源码路径[:-3] + "_en.ul"
        if 文件 == "特殊字符/中文标识符.ul":
            continue

    参数 = [可执行文件, 源码路径]
    进程表[文件] = subprocess.Popen(参数, stdout=subprocess.PIPE)

失败表 = {}

for 文件 in 进程表:
    反馈 = 进程表[文件].communicate()
    输出 = 反馈[0]
    # TODO: 确认报错信息, 现在对语法错误无效
    报错 = 反馈[1]

    预期值 = 期望值[文件]
    if 报错:
        失败表[文件] = 报错
    elif 输出 == 预期值:
        print("通过： " + 文件)
    # elif 报错 == 期望值[文件]:
    #    print("错误一致: " + 文件)
    else:
        失败表[文件] = 输出

print("===================")
if len(失败表) > 0:
    for 文件 in 失败表:
        print(f"失败： {文件} 期望：{预期值} 实际：{失败表[文件]}")
else:
    print("！全部通过！")

print("本测试针对木兰原始可执行文件与重现项目的发布版。在此之前先运行`测试/unittest`下的集成测试。")



import ast


class NameFixPass(ast.NodeTransformer):
    # 只看到如下区别, class 定义中此属性非空: decorator_list=[Name(id='staticmethod', ctx=Load(), lineno=2, col_offset=5)],
    """
    A python NodeVisitor which traverses the generated ast
    to fix the signature of class methods by adding the
    implicit argument 'self' and also convert the function
    name of the class constructors..
    """

    def __init__(self, filename):
        self.filename = filename
        self.类 = ['']

    def visit_FunctionDef(self, 函数):
        if 函数.name.startswith('$'):
            函数.name = 函数.name.replace('$', '')
            if 函数.name == self.类[(-1)]:
                函数.name = '__init__'
            函数.args.args.insert(0, ast.arg(arg='self',
                                             annotation=None,
                                             lineno=(函数.lineno),
                                             col_offset=(函数.col_offset)))
        elif self.类[(-1)]:
            if not 函数.args.args or 函数.args.args[0].arg != 'self':
                decorator = ast.Name(id='staticmethod',
                                     ctx=(ast.Load()),
                                     lineno=(函数.lineno),
                                     col_offset=(函数.col_offset))
                函数.decorator_list.append(decorator)
        self.类.append(None)
        函数 = self.generic_visit(函数)
        self.类.pop(-1)
        return 函数

    def visit_ClassDef(self, 类):
        self.类.append(类.name)
        类 = self.generic_visit(类)
        self.类.pop(-1)
        return 类

class AnnoFuncInsertPass(ast.NodeTransformer):
    """
    Visit all ast to insert each anonymous function
    just before where it has been referenced.
    """

    def __init__(self, 匿名函数):
        self.匿名函数 = 匿名函数

    def generic_visit(self, 节点):
        for 域, 旧值 in ast.iter_fields(节点):
            if isinstance(旧值, list) and len(旧值) > 0:
                if isinstance(旧值[0], ast.stmt):
                    旧值[:] = self.visit_stmts(旧值)
                    continue
                各新值 = []
                for 值 in 旧值:
                    if isinstance(值, ast.AST):
                        值 = self.visit(值)
                        if 值 is None:
                            continue
                        elif not isinstance(值, ast.AST):
                            各新值.extend(值)
                            continue
                    各新值.append(值)

                旧值[:] = 各新值
            elif isinstance(旧值, ast.AST):
                新值 = self.visit(旧值)
                if 新值 is None:
                    delattr(节点, 域)
                else:
                    setattr(节点, 域, 新值)

        return 节点

    def visit_stmts(self, 声明列表):
        新声明列表 = []
        for 声明 in 声明列表:
            声明 = self.visit(声明)
            for 节点 in ast.walk(声明):
                if isinstance(节点, ast.Name) and 节点 in self.匿名函数:
                    新声明列表.append(self.匿名函数[节点])
                    del self.匿名函数[节点]

            新声明列表.append(声明)

        return 新声明列表


from enum import Enum, unique

@unique
class 语法(Enum):
    # 语法名称, 英文为原代码中对应名称. 按定义的先后顺序
    模块 = 'start'
    块 = 'block'
    注水声明列表 = 'stmt_list'
    声明列表 = 'stmt_list_'
    声明 = 'stmt'
    类型定义 = 'type_define'
    各基准类 = 'bases'
    类型主体 = 'type_body'
    各类型内声明 = 'type_stmts'
    类型内声明 = 'type_stmt'
    应变属性 = 'property'
    操作符 = 'operator'
    操作数 = 'op_arg'
    二元操作符 = 'bin_op'
    模块位置 = 'module_name_'
    引用声明 = 'using_stmt'
    各模块名 = 'module_name'
    模块名 = 'module_names'
    表达式声明 = 'expr_stmt'
    表达式前缀 = 'prefix_expr'
    赋值 = 'assignment'
    外部声明 = 'declaration'
    增量赋值 = 'aug(ment)_assign'
    返回声明 = 'return_stmt'
    终止声明 = 'break_stmt'
    跳过声明 = 'continue_stmt'
    试试声明 = 'try_stmt'
    抛出声明 = 'throw_stmt'
    顺便处理 = 'withitem'
    各接手声明 = 'catch_stmts'
    接手声明 = 'catch_stmt'
    表达式 = 'expr'
    各表达式前缀 = 'prefix_exprs'
    片 = 'slice'
    数 = 'number'
    字符串 = 'string'
    列表表达式 = 'list_expr'
    多项式乘法 = 'factor_expr'
    字典表达式 = 'dict_expr'
    各键值对 = 'kv_pairs'
    键值对 = 'kv_pair'
    常量 = 'name_const'
    二元表达式 = 'bin_expr'
    一元表达式 = 'unary_expr'
    三元表达式 = 'ternary_expr'
    首要表达式 = 'primary_expr'
    范围表达式 = 'range_expr'
    调用 = 'call'
    超类 = 'super'
    lambda形参 = 'lambda_param',
    lambda主体 = 'lambda_body',
    lambda表达式 = 'lambda_expr',
    匿名函数 = 'lambda_func'
    类型名称 = 'type_name'
    实参部分 = 'arguments'
    各名称 = 'names'
    各实参 = 'args'
    各表达式 = 'exprs'
    变量 = 'var'
    实参 = 'arg'
    形参列表 = 'param_list'
    非空形参列表 = 'param_list_not_empty'
    形参 = 'param'
    函数 = 'function'
    条件声明 = 'if_stmt'
    否则如果声明 = 'elif_stmt'
    每当声明 = 'while_stmt'
    迭代器 = 'iterator'
    遍历范围 = 'loop_range'
    对于声明 = 'for_stmt'
    名称 = 'name'

    def 成分(self, *成分):
        文本 = []
        for 各成分 in 成分:
            if isinstance(各成分, Enum):
                文本.append(各成分.name)
            else:
                文本.append(各成分)
        return self.name + " : " + " ".join(文本)



import re, sys, traceback
from pathlib import Path

运行时木兰路径 = str(Path("site-packages/木兰/"))

def 反馈信息(例外, 源码文件=None):
    提神符 = " 😰 "
    if sys.platform == 'win32':
        提神符 = "（>﹏<）"
    return 提神符 + "\n".join(中文化(例外, 源码文件))

# TODO: 测试
def 中文化(例外, 源码文件=None):
    类型 = 例外.__class__.__name__
    原信息 = str(例外)
    exc_type, exc_value, 回溯信息 = sys.exc_info()
    各层 = traceback.extract_tb(回溯信息)
    # print(repr(各层))
    各行 = []
    各行.append(提示(类型, 原信息))
    最高层号 = len(各层) - 1
    for 层号 in range(最高层号, -1, -1):
        层 = 各层[层号]
        文件名 = 层.filename
        if 文件名.find(运行时木兰路径) > 0:
            continue

        行信息 = f'见第{层.lineno}行'
        行内容 = 层.line
        if 源码文件 == None and 文件名 == "【标准输入】":
            return [各行[-1] + f"，{行信息}"]
        else:
            # 在第二层前显示
            if 层号 == 最高层号 - 1:
                各行.append("调用层级如下")

            if 文件名 == 源码文件:
                各行.append(行信息 + '：' + 行内容)
            else:
                各行.append(f"“{文件名}”第{层.lineno}行：{行内容}")
    return 各行

def 提示(类型, 原信息):
    if 类型 == 'NameError':
        return re.sub(r"name '(.*)' is not defined", r"请先定义‘\1’再使用", 原信息)
    elif 类型 == 'ZeroDivisionError':
        return "请勿除以零"
    elif 类型 == 'RecursionError':
        return "递归过深。请确认: 1、的确需要递归 2、递归的收敛正确"
    elif 类型 == 'UnboundLocalError':
        return re.sub(
            r"local variable '(.*)' referenced before assignment",
            r"请先对本地变量‘\1’赋值再引用",
            原信息)
    elif 类型 == 'KeyError':
        return "字典中不存在此键：" + 原信息
    elif 类型 == 'TypeError':
        模式 = 'can only concatenate str \(not "(.*)"\) to str'
        if re.match(模式, 原信息):
            return re.sub(模式, r'字符串只能拼接字符串，请将“\1”先用 str() 转换', 原信息)
    elif 类型 == 'IndexError' and 原信息 == "list index out of range":
        return "取列表内容时，索引超出范围"
    elif 类型 == 'AttributeError':
        # TODO: 为 测试/错误处理/属性被静态调用.ul 改进错误信息
        信息 = "需要添加此属性：" + 原信息
        if 原信息 == "__enter__":
            信息 += "\n参考：https://stackoverflow.com/questions/1984325/explaining-pythons-enter-and-exit"
        return 信息
    elif 类型 == 'FileNotFoundError':
        return re.sub(r"\[Errno 2\] No such file or directory: '(.*)'",
            r"没找到文件或路径：‘\1’",
            原信息)
    elif 类型 == 'ModuleNotFoundError':
        return re.sub(r"No module named '(.*)'", r"没找到模块：‘\1’", 原信息)
    return 类型 + "：" + 原信息



'''
Copyright (c) Alex Gaynor and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of rply nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
# 摘自 rply 源码, commit 6e16262dc6d434fc467eed83ed31ca764ba01a34
from rply.errors import ParsingError


class LRParser(object):
    def __init__(self, lrparser):
        self.lr_table = lrparser.lr_table
        self.error_handler = lrparser.error_handler

    def parse(self, tokenizer, state=None):
        from rply.token import Token

        lookahead = None
        lookaheadstack = []

        statestack = [0]
        symstack = [Token("$end", "$end")]

        current_state = 0
        while True:
            if self.lr_table.default_reductions[current_state]:
                t = self.lr_table.default_reductions[current_state]
                current_state = self._reduce_production(
                    t, symstack, statestack, state
                )
                continue

            if lookahead is None:
                if lookaheadstack:
                    lookahead = lookaheadstack.pop()
                else:
                    try:
                        lookahead = next(tokenizer)
                    except StopIteration:
                        lookahead = None

                if lookahead is None:
                    lookahead = Token("$end", "$end")

            ltype = lookahead.gettokentype()
            if ltype in self.lr_table.lr_action[current_state]:
                t = self.lr_table.lr_action[current_state][ltype]
                if t > 0:
                    statestack.append(t)
                    current_state = t
                    symstack.append(lookahead)
                    lookahead = None
                    continue
                elif t < 0:
                    current_state = self._reduce_production(
                        t, symstack, statestack, state
                    )
                    continue
                else:
                    n = symstack[-1]
                    return n
            else:
                # TODO: actual error handling here
                if self.error_handler is not None:
                    if state is None:
                        self.error_handler(lookahead)
                    else:
                        self.error_handler(state, lookahead)
                    lookahead = None
                    continue
                else:
                    raise ParsingError(None, lookahead.getsourcepos())

    def _reduce_production(self, t, symstack, statestack, state):
        # reduce a symbol on the stack and emit a production
        p = self.lr_table.grammar.productions[-t]
        pname = p.name
        plen = p.getlength()
        start = len(symstack) + (-plen - 1)
        assert start >= 0
        targ = symstack[start + 1:]
        start = len(symstack) + (-plen)
        assert start >= 0
        del symstack[start:]
        del statestack[start:]
        if state is None:
            value = p.func(targ)
        else:
            value = p.func(state, targ)
        symstack.append(value)
        current_state = self.lr_table.lr_goto[statestack[-1]][pname]
        statestack.append(current_state)
        return current_state



import unittest
import ast
from rply import LexingError

from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.分析器.词法分析器 import 分词器
from 木兰.分析器.错误 import 语法错误
from 编辑器.运行木兰 import 运行木兰代码

# TODO：需确保无此类 Warning：ParserGeneratorWarning: 28 shift/reduce conflicts
class test语法树(unittest.TestCase):

    def test_词法分析(self):
        # 由于换行也包括在内，"后括号"一词的位置并非 2：1
        各词 = self.分词("{print(2)\n}")
        for 词 in 各词:
            最末 = 词
        self.assertEqual(最末.getstr(), "\n}")
        self.assertEqual(最末.getsourcepos().lineno, 1)
        self.assertEqual(最末.getsourcepos().colno, 10)

    def test_行列号(self):
        节点 = self.生成语法树("print(1/0)")
        expr节点 = self.取子节点(节点, "body", 0)
        call节点 = self.取子节点(expr节点, "value")
        self.assertEqual(call节点.lineno, 1)
        self.assertEqual(call节点.col_offset, 1)

        除法节点 = self.取子节点(call节点, "func")
        self.assertEqual(除法节点.lineno, 1)
        self.assertEqual(除法节点.col_offset, 1)

    # 必须在 Python 3.7 下运行. 注意：木兰源码勿带多余空格，否则行列数不一致。
    def test_整树比较(self):
        路径 = "测试/unittest/例程/"
        期望值 = {
            "反斜杠.ul": r"Module(body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=1, col_offset=1), args=[Str(s='\\', lineno=1, col_offset=7)], keywords=[], lineno=1, col_offset=1), lineno=1, col_offset=1)])",
            "带类型参数.ul": "Module(body=[FunctionDef(name='hello', args=arguments(args=[arg(arg='p', annotation=Name(id='Person', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Num(n=1, lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1)])",
            "类型.ul": "Module(body=[ClassDef(name='Person', bases=[], keywords=[], body=[FunctionDef(name='__add__', args=arguments(args=[arg(arg='self', lineno=2, col_offset=12), arg(arg='a', annotation=None, lineno=2, col_offset=13)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Call(func=Name(id='Person', ctx=Load(), lineno=2, col_offset=25), args=[BinOp(left=Attribute(value=Name(id='self', ctx=Load(), lineno=2, col_offset=32), attr='name', ctx=Load(), lineno=2, col_offset=32), op=Add(), right=Attribute(value=Name(id='a', ctx=Load(), lineno=2, col_offset=44), attr='name', ctx=Load(), lineno=2, col_offset=44), lineno=2, col_offset=32)], keywords=[], lineno=2, col_offset=25), lineno=2, col_offset=18)], decorator_list=[], lineno=2, col_offset=1), FunctionDef(name='__init__', args=arguments(args=[arg(arg='self', annotation=None, lineno=3, col_offset=1), arg(arg='name', annotation=None, lineno=3, col_offset=14)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Assign(targets=[Attribute(value=Name(id='self', ctx=Load(), lineno=3, col_offset=22), attr='name', ctx=Store(), lineno=3, col_offset=22)], value=Name(id='name', ctx=Load(), lineno=3, col_offset=34), lineno=3, col_offset=22)], decorator_list=[], lineno=3, col_offset=1)], decorator_list=[], lineno=1, col_offset=1)])",
            "函数.ul": "Module(body=[FunctionDef(name='echo', args=arguments(args=[arg(arg='number', annotation=None, lineno=1, col_offset=11)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='number', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], decorator_list=[], lineno=1, col_offset=1), Expr(value=Call(func=Name(id='echo', ctx=Load(), lineno=4, col_offset=1), args=[Num(n=2, lineno=4, col_offset=6)], keywords=[], lineno=4, col_offset=1), lineno=4, col_offset=1)])",
            "赋值.ul": "Module(body=[Assign(targets=[Name(id='a', ctx=Store(), lineno=1, col_offset=1)], value=Num(n=2, lineno=1, col_offset=3), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='a', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)])",
            "多行运算.ul": "Module(body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=1, col_offset=1), args=[Num(n=0, lineno=1, col_offset=7)], keywords=[], lineno=1, col_offset=1), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[BinOp(left=BinOp(left=Num(n=1, lineno=2, col_offset=7), op=Add(), right=BinOp(left=Num(n=2, lineno=2, col_offset=9), op=Mult(), right=Num(n=3, lineno=2, col_offset=11), lineno=2, col_offset=9), lineno=2, col_offset=7), op=Sub(), right=Call(func=Name(id='__div__', ctx=Load(), lineno=2, col_offset=13), args=[Num(n=4, lineno=2, col_offset=13), Num(n=5, lineno=2, col_offset=15)], keywords=[], lineno=2, col_offset=13), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)])"
        }
        for 文件 in 期望值:
            源码路径 = 路径 + 文件
            节点 = self.读源码生成树(源码路径)
            self.assertEqual(ast.dump(节点, True, True), 期望值[文件], f"\"{文件}\"出错")

    def test_报错(self):
        try:
            节点 = self.生成语法树("using func")
            self.fail("不该到这")
        except 语法错误 as e:
            self.assertEqual(e.列号, 7)

    def test_词不识(self):
        try:
            self.读源码生成树("测试/错误处理/词不识.ul")
        except 语法错误 as e:
            self.assertEqual(e.信息, "分词时没认出这个词 \"#\"")
            self.assertEqual(e.行号, 3, "注意：此用例需安装包含此 commit 的 rply：https://github.com/alex/rply/commit/6e16262dc6d434fc467eed83ed31ca764ba01a34")
            self.assertEqual(e.列号, 6) # 应该为 1, 在rply 提问: https://github.com/alex/rply/pull/95#issuecomment-729513800

    def test_try随意(self):
        self.assertTrue(运行木兰代码("测试/错误处理/try随意.ul").find("要添加此属性：__enter__") > 0)

    def 分词(self, 源码):
        return 分词器.lex(源码)

    def 生成语法树(self, 源码):
        分析器 = 语法分析器()
        return 分析器.分析(源码, '')

    def 读源码生成树(self, 源码文件):
        with open(源码文件, 'r', encoding='utf-8') as f:
            源码 = f.read()
            return self.生成语法树(源码)

    def 取子节点(self, 节点, 子节点名, 索引 = -1):
        for 子节点 in ast.iter_fields(节点):
            if 子节点[0] == 子节点名:
                if isinstance(子节点[1], list):
                    return 子节点[1][索引]
                else:
                    return 子节点[1]

if __name__ == '__main__':
    unittest.main()


import unittest
from 木兰.交互 import 括号已配对

class test交互(unittest.TestCase):

    def test_括号已配对(self):
        self.assertEqual(括号已配对("a=3\n"), True)
        self.assertEqual(括号已配对("func a() {}\n"), True)
        self.assertEqual(括号已配对("if true\n\n"), True)
        self.assertEqual(括号已配对("if true\n;\n"), True)

        self.assertEqual(括号已配对("1+\\n"), False)
        self.assertEqual(括号已配对("if\n"), False)
        self.assertEqual(括号已配对("elif\n"), False)
        self.assertEqual(括号已配对("else\n"), False)
        self.assertEqual(括号已配对("for\n"), False)
        self.assertEqual(括号已配对("func\n"), False)
        self.assertEqual(括号已配对("func a\n"), False)
        self.assertEqual(括号已配对("func a(\n"), False)
        self.assertEqual(括号已配对("func a()\n"), False)
        self.assertEqual(括号已配对("func a() {\n"), False)
        self.assertEqual(括号已配对("loop\n"), False)
        self.assertEqual(括号已配对("operator\n"), False)
        self.assertEqual(括号已配对("type\n"), False)
        self.assertEqual(括号已配对("while\n"), False)

if __name__ == '__main__':
    unittest.main()


from enum import Enum, unique
from rply import ParserGenerator
from rply.errors import LexingError
from 木兰.分析器.rply_parser import LRParser
from 木兰.分析器.语法树 import *
from 木兰.分析器.错误 import 语法错误, 词法错误

from 木兰.分析器.词法分析器 import *
from 木兰.分析器.语法树处理 import *
from 木兰.分析器.语法成分 import *

from 木兰.功用.常用 import *

"""
LR(1) 将木兰源码分析后生成 Python 语法树
"""
class 语法分析器:

    # TODO: 改进可视化 parse 过程(各个语法规则的顺序), 方便调试
    调试 = False

    分析器母机 = ParserGenerator(
        规则,
        precedence=[
            #('nonassoc', ['标识符']),
            ('right', [箭头]),
            ('right', [问号, 冒号]),
            ('left', [连词_或]),
            ('left', [连词_且]),
            # nonassoc 参考: http://www.dabeaz.com/ply/ply.html
            # non-associativity in the precedence table. This would be used when you don't want operations to chain together
            ('nonassoc', [大于, 小于, 大于等于, 小于等于, 严格不等于, 严格等于]),
            ('left', [不等于, 等于]),
            ('nonassoc', [连词_每隔]),
            ('nonassoc', [点点, 点点小于]),
            #('nonassoc', ['(']),
            ('left', [加, 减]),
            ('left', [星号, 除, 求余]),
            ('left', [非]),
            ('right', [乘方])
        ]
    )

    # ast 参考: https://docs.python.org/3.7/library/ast.html#abstract-grammar

    @分析器母机.production(语法.模块.成分(语法.注水声明列表))
    def 模块(self, 片段):
        return 语法树.新节点(语法.模块, 主体=片段[0], 忽略类型=[])

    @分析器母机.production(语法.块.成分(分号))
    @分析器母机.production(语法.块.成分(前大括号, 语法.注水声明列表, 后大括号))
    def 块(self, 片段):
        if 语法分析器.调试:
            print('块')
        if len(片段) == 3:
            if 片段[1]:
                return 片段[1]
        return [语法树.空转(片段)]

    @分析器母机.production(语法.注水声明列表.成分())
    @分析器母机.production(语法.注水声明列表.成分(语法.声明列表))
    @分析器母机.production(语法.注水声明列表.成分(语法.声明列表, 换行))
    @分析器母机.production(语法.注水声明列表.成分(语法.声明列表, 分号))
    def 注水声明列表(self, 片段):
        if 语法分析器.调试:
            print('注水声明列表')
        if len(片段) > 0:
            return 片段[0]
        return []

    @分析器母机.production(语法.声明列表.成分(语法.声明))
    @分析器母机.production(语法.声明列表.成分(语法.声明列表, 换行, 语法.声明))
    @分析器母机.production(语法.声明列表.成分(语法.声明列表, 分号, 语法.声明))
    def 声明列表(self, 片段):
        if 语法分析器.调试:
            print('声明列表')
        if len(片段) == 1:
            return [片段[0]]
        片段[0].append(片段[(-1)])
        return 片段[0]

    @分析器母机.production(语法.声明.成分(语法.类型定义))
    @分析器母机.production(语法.声明.成分(语法.函数))
    @分析器母机.production(语法.声明.成分(语法.条件声明))
    @分析器母机.production(语法.声明.成分(语法.每当声明))
    @分析器母机.production(语法.声明.成分(语法.对于声明))
    @分析器母机.production(语法.声明.成分(语法.外部声明))
    def 混合声明(self, 片段):
        if 语法分析器.调试:
            print('混合声明')
        return 片段[0]

    @分析器母机.production(语法.类型定义.成分(名词_类型, 语法.名称, 语法.各基准类, 语法.类型主体))
    def 类型定义(self, 片段):
        return 语法树.新节点(语法.类型定义,
            名称=片段[1].id,
            各基准类=片段[2],
            主体=片段[-1],
            片段=片段)

    @分析器母机.production(语法.各基准类.成分())
    @分析器母机.production(语法.各基准类.成分(冒号, 语法.表达式前缀))
    @分析器母机.production(语法.各基准类.成分(冒号, 语法.各表达式前缀))
    def 各基准类(self, 片段):
        if 语法分析器.调试:
            print('各基准类')
        if len(片段) == 0:
            return []
        if isinstance(片段[1], list):
            return 片段[1]
        return [片段[1]]

    @分析器母机.production(语法.类型主体.成分(前大括号, 语法.各类型内声明, 后大括号))
    def 类型主体(self, 片段):
        if 语法分析器.调试:
            print('类型主体')
        return 片段[1]

    @分析器母机.production(语法.各类型内声明.成分())
    @分析器母机.production(语法.各类型内声明.成分(语法.各类型内声明, 语法.类型内声明))
    def 各类型内声明(self, 片段):
        if 语法分析器.调试:
            print('各类型内声明')
        if len(片段) == 0:
            # TODO: 应该报语法错误?
            return []
        if isinstance(片段[1], list): # TODO: 有这种情况吗?
            片段[0] += 片段[1]
        else:
            片段[0].append(片段[1])
        return 片段[0]

    @分析器母机.production(语法.类型内声明.成分(语法.块))
    @分析器母机.production(语法.类型内声明.成分(语法.操作符))
    @分析器母机.production(语法.类型内声明.成分(语法.函数))
    @分析器母机.production(语法.类型内声明.成分(语法.应变属性))
    def 类型内声明(self, 片段):
        if 语法分析器.调试:
            print('类型内声明')
        return 片段[0]

    @分析器母机.production(语法.应变属性.成分(名词_应变属性, 标识符, 语法.块))
    @分析器母机.production(语法.应变属性.成分(名词_应变属性, 标识符, 前小括号, 后小括号, 语法.块))
    @分析器母机.production(语法.应变属性.成分(名词_应变属性, 标识符, 符号_赋值, 前小括号, 语法.形参, 后小括号, 语法.块))
    def 应变属性(self, 片段):
        名称 = 片段[1].getstr()
        函数 = 语法树.新节点(语法.函数, 名称=名称,
                        参数=self.形参列表(),
                        主体=片段[-1],
                        片段=片段)
        if len(片段) == 7:
            函数.args.args.append(片段[4])
            函数.decorator_list.append(
                语法树.属性(
                    值=语法树.新节点(语法.名称,
                        标识=名称.replace('$', ''),
                        上下文=ast.Load(),
                        片段=片段),
                    属性='setter',
                    片段=片段))
        else:
            函数.decorator_list.append(
                语法树.新节点(语法.名称,
                    标识='property',
                    上下文=ast.Load(),
                    片段=片段))
        return 函数

    @分析器母机.production(语法.操作符.成分(名词_操作符, 语法.二元操作符, 语法.操作数, 语法.块))
    def 操作符(self, 片段):
        return 语法树.新节点(语法.函数, 名称=片段[1],
                        参数=片段[2],
                        主体=片段[-1],
                        片段=片段)

    @分析器母机.production(语法.操作数.成分(前小括号, 语法.形参, 后小括号))
    def 操作数(self, 片段):
        各形参 = self.形参列表()
        各形参.args.append(语法树.新节点(语法.操作数, 参数='self',
                               片段=片段))
        [各形参.args.append(形参) for 形参 in 片段 if isinstance(形参, ast.arg)]
        return 各形参

    @分析器母机.production(语法.二元操作符.成分(加))
    @分析器母机.production(语法.二元操作符.成分(减))
    @分析器母机.production(语法.二元操作符.成分(等于))
    def 二元操作符(self, 片段):
        对照表 = {
            加: '__add__',
            减: '__sub__',
            等于: '__eq__',
        }
        return 对照表[片段[0].getstr()]

    @分析器母机.production(语法.声明.成分(语法.引用声明))
    @分析器母机.production(语法.声明.成分(语法.表达式声明))
    @分析器母机.production(语法.声明.成分(语法.赋值))
    @分析器母机.production(语法.声明.成分(语法.增量赋值))
    @分析器母机.production(语法.声明.成分(语法.终止声明))
    @分析器母机.production(语法.声明.成分(语法.跳过声明))
    @分析器母机.production(语法.声明.成分(语法.试试声明))
    @分析器母机.production(语法.声明.成分(语法.抛出声明))
    @分析器母机.production(语法.声明.成分(语法.返回声明))
    def 声明(self, 片段):
        return 片段[0]

    # TODO: 补完 try-catch-throw

    @分析器母机.production(语法.抛出声明.成分(动词_抛出, 语法.表达式))
    def 抛出声明(self, 片段):
        return 语法树.引发(片段[1], 片段)

    @分析器母机.production(语法.顺便处理.成分(语法.表达式前缀, 符号_赋值, 语法.表达式))
    @分析器母机.production(语法.顺便处理.成分(语法.表达式))
    def 顺便处理(self, 片段):
        处理项 = 语法树.顺便处理项(上下文表达式=片段[(-1)], 可选变量=None, 片段=片段)
        if len(片段) == 3:
            片段[0].ctx = ast.Store()
            处理项.optional_vars = 片段[0]
        return 处理项

    @分析器母机.production(语法.试试声明.成分(动词_试试, 语法.顺便处理, 语法.块))
    def 试试声明_顺便处理(self, 片段):
        return 语法树.顺便(各项=[片段[1]], 主体=片段[2], 片段=片段)

    @分析器母机.production(语法.试试声明.成分(动词_试试, 语法.块, 语法.各接手声明))
    def 试试声明(self, 片段):
        return 语法树.试试(主体=(片段[1]),
                      处理=片段[-1],
                      片段=片段)

    @分析器母机.production(语法.各接手声明.成分(语法.各接手声明, 语法.接手声明))
    @分析器母机.production(语法.各接手声明.成分(语法.接手声明))
    def 各接手声明(self, 片段):
        if len(片段) == 1:
            return [片段[0]]
        片段[0].append(片段[1])
        return 片段[0]

    @分析器母机.production(语法.接手声明.成分(动词_接手, 语法.名称, 冒号, 语法.表达式, 语法.块))
    @分析器母机.production(语法.接手声明.成分(动词_接手, 语法.名称, 语法.块))
    def 接手声明(self, 片段):
        名称, 类型 = (None, None)
        if len(片段) == 3:
            名称 = 片段[1].id
        elif len(片段) == 5:
            名称 = 片段[1].id
            类型 = 片段[3]
        return 语法树.例外处理(类型=类型,
                        名称=名称,
                        主体=片段[-1],
                        片段=片段)

    # TODO: 更多引用方式

    @分析器母机.production(语法.模块位置.成分(语法.模块名))
    @分析器母机.production(语法.模块位置.成分(点))
    def 模块位置(self, 片段):
        return 片段[0]

    @分析器母机.production(语法.引用声明.成分(动词_引用, 语法.各模块名, 连词_于, 语法.模块位置))
    @分析器母机.production(语法.引用声明.成分(动词_引用, 星号, 连词_于, 语法.模块位置))
    def 引用于(self, 片段):
        模块, 层 = 片段[-1], 0
        if isinstance(片段[-1], Token):
            模块 = None
            层 = 1 if 片段[-1].getstr() == '.' else 2
        节点 = 语法树.从模块导入(
            模块=模块,
            各名称=[],
            位置=层,
            片段=片段)
        if isinstance(片段[1], list):
            节点.names += 片段[1]
        else:
            节点.names.append(
                语法树.别名(名称=星号,
                        别名=None,
                        片段=片段[1]))
        return 节点

    @分析器母机.production(语法.引用声明.成分(动词_引用, 语法.各模块名))
    def 引用声明(self, 片段):
        return 语法树.新节点(语法.引用声明,
            名称=片段[1],
            片段=片段)

    @分析器母机.production(语法.各模块名.成分(语法.模块名))
    @分析器母机.production(语法.各模块名.成分(语法.各模块名, 逗号, 语法.模块名))
    def 各模块名(self, 片段):
        别名 = 语法树.别名(
            名称=片段[-1],
            别名=None,
            片段=片段[-1]
        )
        if len(片段) == 1:
            return [别名]
        片段[0].append(别名)
        return 片段[0]

    @分析器母机.production(语法.模块名.成分(语法.模块名, 点, 语法.名称))
    @分析器母机.production(语法.模块名.成分(语法.名称))
    def 模块名(self, 片段):
        if len(片段) == 1:
            return 片段[0].id
        return '%s.%s' % (片段[0], 片段[2].id)

    @分析器母机.production(语法.表达式声明.成分(语法.表达式前缀))
    def 表达式声明(self, 片段):
        if 语法分析器.调试:
            print("表达式声明")
        if not isinstance(片段[0], ast.Call):
            # TODO：下面两个似乎不需要
            # if not isinstance(片段[0], ast.Yield):
                #if not isinstance(片段[0], ast.Str):
                    片段[0] = 语法树.新节点(语法.调用, 函数=片段[0],
                      参数=[],
                      关键词=[],
                      片段=片段)
        return 语法树.新节点(语法.表达式, 值 = 片段[0], 片段 = 片段)

    @分析器母机.production(语法.返回声明.成分(动词_返回))
    @分析器母机.production(语法.返回声明.成分(动词_返回, 语法.各表达式))
    def 返回(self, 片段):
        值 = None
        if len(片段) == 2:
            if len(片段[1]) == 1:
                值 = 片段[1][0]
            else:
                值 = 语法树.多项(元素=片段[1],
                           上下文=ast.Load(),
                           片段=片段[1])
        return 语法树.新节点(语法.返回声明,
            值=值,
            片段=片段)

    @分析器母机.production(语法.增量赋值.成分(语法.表达式前缀, 增量赋值, 语法.表达式))
    @分析器母机.production(语法.增量赋值.成分(语法.表达式前缀, 减量赋值, 语法.表达式))
    def 增量赋值(self, 片段):
        运算符 = 片段[1].getstr()
        对照表 = {
            增量赋值: ast.Add(),
            减量赋值: ast.Sub(),
        }
        if 运算符 in 对照表:
            python运算 = 对照表[运算符]

        # 否则报错：
        # ValueError: expression must have Store context but has Load instead
        片段[0].ctx = ast.Store()

        return 语法树.新节点(语法.增量赋值, 变量=片段[0], 运算符=python运算, 值=片段[2], 片段=片段)

    @分析器母机.production(语法.赋值.成分(语法.表达式前缀, 符号_赋值, 语法.表达式))
    def 赋值(self, 片段):
        if 语法分析器.调试:
            print("赋值")
        片段[0].ctx = ast.Store()
        return 语法树.新节点(语法.赋值,
            变量 = 片段[0],
            值 = 片段[2],
            片段 = 片段)

    @分析器母机.production(语法.外部声明.成分(形容词_外部, 语法.各名称))
    def 外部声明(self, 片段):
        return 语法树.新节点(语法.外部声明, 名称=[名称.id for 名称 in 片段[1]], 片段=片段)

    @分析器母机.production(语法.各表达式前缀.成分(语法.各表达式前缀, 逗号, 语法.表达式前缀))
    @分析器母机.production(语法.各表达式前缀.成分(语法.表达式前缀, 逗号, 语法.表达式前缀))
    def 各表达式前缀(self, 片段):
        if isinstance(片段[0], list):
            片段[0].append(片段[2])
            return 片段[0]
        return [片段[0], 片段[2]]

    @分析器母机.production(语法.赋值.成分(语法.各表达式前缀, 符号_赋值, 语法.各表达式))
    def 多项赋值(self, 片段):
        for 表达式前缀 in 片段[0]:
            表达式前缀.ctx = ast.Store()

        左边 = 语法树.多项(
            元素=片段[0],
            上下文=ast.Store(),
            片段=片段)
        if len(片段[2]) > 1:
            右边 = 语法树.多项(
                元素=片段[2],
                上下文=ast.Load(),
                片段=片段[2])
        else:
            右边 = 片段[2][0]
        return 语法树.新节点(语法.赋值,
            变量=左边,
            值=右边,
            片段=片段)

    @分析器母机.production(语法.终止声明.成分(动词_终止))
    def 终止声明(self, 片段):
        return 语法树.新节点(语法.终止声明, 片段 = 片段)

    @分析器母机.production(语法.跳过声明.成分(动词_跳过))
    def 跳过声明(self, 片段):
        return 语法树.新节点(语法.跳过声明, 片段 = 片段)

    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 加, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 减, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 星号, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 乘方, 语法.表达式))
    def 二元表达式(self, 片段):
        左 = 片段[0]
        右 = 片段[2]
        运算符 = 片段[1].getstr()
        对照表 = {
            加: ast.Add(),
            减: ast.Sub(),
            星号: ast.Mult(),
            乘方: ast.Pow()
        }
        if 运算符 in 对照表:
            python运算 = 对照表[运算符]
        else:
            breakpoint()
        return 语法树.新节点(语法.二元表达式, 左=左, 运算符=python运算, 右=右, 片段=片段)

    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 除, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 求余, 语法.表达式))
    def 除法(self, 片段):
        操作符 = '__div__' if 片段[1].getstr() == '/' else '__rem__'
        return 语法树.新节点(语法.调用,
            函数=语法树.新节点(语法.名称,
                标识=操作符,
                上下文=(ast.Load()),
                片段=片段),
            参数=[片段[0], 片段[2]],
            关键词=[],
            片段=片段)

    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 大于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 大于等于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 小于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 小于等于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 等于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 不等于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 严格等于, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 严格不等于, 语法.表达式))
    def 比较(self, 片段):
        对照表 = {
            大于: ast.Gt(),
            大于等于: ast.GtE(),
            小于: ast.Lt(),
            小于等于: ast.LtE(),
            等于: ast.Eq(),
            不等于: ast.NotEq(),
            严格等于: ast.Is(),
            严格不等于: ast.IsNot()
            }
        return 语法树.新节点(语法.二元表达式,
            前项 = 片段[0],
            运算符 = 对照表[片段[1].getstr()],
            后项 = 片段[2],
            片段=片段)

    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 连词_且, 语法.表达式))
    @分析器母机.production(语法.二元表达式.成分(语法.表达式, 连词_或, 语法.表达式))
    def 布尔表达式(self, 片段):
        return 语法树.新节点(语法.二元表达式,
            运算符=(ast.And() if 片段[1].getstr() == 连词_且 else ast.Or()),
            前项 = 片段[0],
            后项 = 片段[2],
            片段=片段)

    @分析器母机.production(语法.范围表达式.成分(语法.表达式, 点点, 语法.表达式))
    @分析器母机.production(语法.范围表达式.成分(语法.表达式, 点点小于, 语法.表达式))
    @分析器母机.production(语法.范围表达式.成分(语法.范围表达式, 连词_每隔, 语法.表达式))
    def 范围表达式(self, 片段):
        连词 = 片段[1].getstr()
        if 连词 != 连词_每隔:
            起 = 片段[0]
            止 = 片段[2]
            if 连词 == 点点:
                止 = 语法树.新节点(语法.二元表达式, 左=止,
                            运算符=ast.Add(),
                            右=语法树.新节点(语法.数, 值=1, 片段=止),
                            片段=止)
                止.fixed = True
            return 语法树.新节点(语法.调用,
                函数=语法树.新节点(语法.名称, 标识='range',
                          上下文=ast.Load(),
                          片段=片段),
                参数=[起, 止],
                关键词=[],
                片段=片段)
        assert isinstance(片段[0], ast.Call)    # 第三个格式
        参数 = 片段[0].args

        # 不允许多个 by 串联
        if len(参数) == 3:
            raise 语法错误(
                信息=('没认出这个词 "%s"' % 片段[1].getstr()),
                文件名=self.文件名,
                行号=语法树.取行号(片段[1]),
                列号=语法树.取列号(片段[1]),
                源码=self.源码)
        参数.append(片段[2])
        止 = 参数[1]
        if hasattr(止, 'fixed'):
            if isinstance(片段[2], ast.Num):
                if 片段[2].n < 0:
                    止.op = ast.Sub()
            else:
                为正 = 语法树.新节点(语法.二元表达式,
                        前项=片段[2],
                        运算符=ast.Gt(),
                        后项=语法树.新节点(语法.数, 值=0, 片段=片段[2]),
                        片段=片段[2])
                增量 = 语法树.如果表达式(
                    条件=为正,
                    主体=止.right,
                    否则=语法树.新节点(语法.数, 值=-1, 片段=片段[2]),
                    片段=片段)
                止.right = 增量
            delattr(止, 'fixed')
        return 片段[0]

    # TODO: #
    @分析器母机.production(语法.一元表达式.成分(减, 语法.表达式))
    @分析器母机.production(语法.一元表达式.成分(非, 语法.表达式))
    @分析器母机.production(语法.一元表达式.成分(取反, 语法.表达式), precedence=非)
    def 一元表达式(self, 片段):
        操作符 = 片段[0].getstr()
        对照表 = {
            减: ast.USub(),
            非: ast.Not(),
            取反: ast.Invert(),
        }
        return 语法树.新节点(语法.一元表达式, 运算符=对照表[操作符], 值=片段[1], 片段=片段)

    @分析器母机.production(语法.三元表达式.成分(语法.表达式, 问号, 语法.表达式, 冒号, 语法.表达式))
    def 三元表达式(self, 片段):
        return 语法树.如果表达式(
            条件=片段[0],
            主体=片段[2],
            否则=片段[-1],
            片段=片段)

    @分析器母机.production(语法.首要表达式.成分(前小括号, 语法.表达式, 后小括号))
    def 首要表达式(self, 片段):
        return 片段[1]

    @分析器母机.production(语法.表达式前缀.成分(语法.调用))
    @分析器母机.production(语法.表达式前缀.成分(语法.变量))
    @分析器母机.production(语法.表达式前缀.成分(语法.匿名函数))
    @分析器母机.production(语法.表达式前缀.成分(语法.字符串))
    @分析器母机.production(语法.表达式前缀.成分(语法.列表表达式))
    @分析器母机.production(语法.表达式前缀.成分(语法.字典表达式))
    def 表达式前缀(self, 片段):
        if 语法分析器.调试:
            print("表达式前缀")
        return 片段[0]

    @分析器母机.production(语法.片.成分(语法.表达式))
    def 索引(self, 片段):
        return 语法树.索引(值=片段[0], 片段=片段)

    @分析器母机.production(语法.片.成分(语法.表达式, 冒号, 语法.表达式))
    @分析器母机.production(语法.片.成分(语法.表达式, 冒号))
    @分析器母机.production(语法.片.成分(冒号, 语法.表达式))
    @分析器母机.production(语法.片.成分(冒号))
    def 片表示(self, 片段):
        下限, 上限 = 片段[0], 片段[-1]
        if isinstance(下限, Token):
            下限 = None
        if isinstance(上限, Token):
            上限 = None
        return 语法树.片(下限, 上限, 片段)

    @分析器母机.production(语法.变量.成分(语法.表达式前缀, 前中括号, 语法.片, 后中括号))
    def 片表达式(self, 片段):
        return 语法树.下标(片段[0], 片段[2], 片段)

    @分析器母机.production(语法.变量.成分(语法.表达式前缀, 点, 语法.名称))
    def 属性表达式(self, 片段):
        return 语法树.属性(
            值=片段[0],
            属性=片段[2].id,
            片段=片段)

    @分析器母机.production(语法.变量.成分(语法.名称))
    def 变量(self, 片段):
        return 片段[0]

    @分析器母机.production(语法.实参部分.成分(前小括号, 语法.各实参, 后小括号))
    @分析器母机.production(语法.实参部分.成分(前小括号, 后小括号))
    def 实参部分(self, 片段):
        if 语法分析器.调试:
            print("实参部分")
        if len(片段) != 3:
            return []
        return 片段[1]

    # 变量->表达式前缀, 以支持高阶函数调用, 如 f(1)(2)
    @分析器母机.production(语法.调用.成分(语法.表达式前缀, 语法.实参部分))
    def 调用(self, 片段):
        if 语法分析器.调试:
            print("调用")
        各参数 = []
        关键词 = []
        for 值, 键 in 片段[1]:
            if 键 is None:
                各参数.append(值)
            else:
                关键词.append(ast.keyword(arg=键,value=值))

        return 语法树.新节点(语法.调用,
                函数=片段[0],
                参数=各参数,
                关键词=关键词,
                片段=片段)

    # TODO: 补全 SUPER
    @分析器母机.production(语法.调用.成分(语法.超类))
    def 调用超类(self, 片段):
        return 片段[0]

    @分析器母机.production(语法.超类.成分(名词_超类))
    def 超类(self, 片段):
        超类节点 = [语法树.新节点(语法.名称, 标识='super',
            上下文=ast.Load(),
            片段=片段)]
        超类节点.append([])
        return self.调用(超类节点)

    @分析器母机.production(语法.lambda形参.成分(语法.名称))
    def lambda形参(self, 片段):
        if isinstance(片段[0], ast.Name):
            args = 语法树.新节点(语法.形参列表, 参数=[],
                片段=片段)
            arg = 语法树.新节点(语法.lambda形参, 参数=片段[0].id,
                标注=None,
                片段=片段)
            args.args = [arg]
            return args

        # TODO：添加测试
        raise SyntaxError(message='expect an identifier here',
            filename=(self.文件名),
            lineno=(self.getlineno(片段)),
            colno=(self.getcolno(片段)),
            source=(self.源码))

    @分析器母机.production(语法.lambda主体.成分(箭头, 语法.表达式))
    @分析器母机.production(语法.lambda主体.成分(箭头, 语法.块))
    def lambda主体(self, 片段):
        return 片段[1]

    @分析器母机.production(语法.lambda表达式.成分(语法.lambda形参, 语法.lambda主体))
    def lambda表达式(self, 片段):
        #if len(p) == 1:
        #    return p[0]
        if isinstance(片段[-1], ast.expr):
            return 语法树.新节点(语法.lambda表达式, 参数=片段[0],
              主体=片段[-1],
              片段=片段)
        return self.创建匿名函数(片段, 片段[0], 片段[-1])

    @分析器母机.production(语法.匿名函数.成分(名词_函数, 前小括号, 语法.形参列表, 后小括号, 语法.块))
    @分析器母机.production(语法.匿名函数.成分(名词_函数, 语法.块))
    def lambda_func(self, 片段):
        形参 = 片段[2] if len(片段) == 5 else [] # lambda_无参.ul 测试错误。如果改为`self.形参列表()`则输出 4
        return self.创建匿名函数(片段, 形参, 片段[-1])

    @分析器母机.production(语法.匿名函数.成分(名词_函数, 前小括号, 语法.形参列表, 后小括号, 冒号, 语法.类型名称, 语法.块))
    def 指定类型匿名函数(self, 片段):
        形参, 主体, 返回 = 片段[2], 片段[-1], 片段[-2]
        return self.创建匿名函数(片段, 形参, 主体, 返回)

    @分析器母机.production(语法.类型名称.成分(语法.名称))
    def 类型名称(self, 片段):
        return 片段[0]

    def 创建匿名函数(self, 片段, 形参, 主体, 返回=None):
        函数名 = 语法树.新节点(语法.名称, 标识=随机文本(),
            上下文=ast.Load(),
            片段=片段)
        函数 = 语法树.新节点(语法.函数, 名称=函数名.id,
                        参数=形参,
                        主体=主体,
                        返回=返回,
                        片段=片段)
        self.匿名函数[函数名] = 函数
        return 函数名

    @分析器母机.production(语法.数.成分(整数))
    @分析器母机.production(语法.数.成分(小数))
    def 数(self, 片段):
        try:
            return 语法树.新节点(语法.数, 值=int(片段[0].getstr(), 0), 片段=片段)
        except ValueError:
            return 语法树.新节点(语法.数, 值=float(片段[0].getstr()), 片段=片段)

    @分析器母机.production(语法.字符串.成分(双引号字符串))
    @分析器母机.production(语法.字符串.成分(单引号字符串))
    def 字符串(self, 片段):
        值 = 片段[0].getstr()

        # TODO: 其他转义字符，如 () 等等
        值 = 值.replace(r'\n', '\n').replace(r'\t', '\t').replace(r'\\', '\\')
        if 值.startswith('"'):
            值 = 值.replace(r'\"', '"')
        else:
            值 = 值.replace(r"\'", "'")
        值 = 值[1:-1]
        插值后 = self.执行插值(值, 片段)
        if 插值后 is not None:
            return 插值后
        return 语法树.新节点(语法.字符串, 值=值, 片段=片段)

    def 执行插值(self, 原始字符串, 片段):
        import re
        模式 = re.compile(r'\\\(([^\\\)]*)\\\)|`([^`]*)`')
        占位替代 = 原始字符串
        所有插值 = []
        位置 = 0
        while True:
            匹配 = 模式.search(string=原始字符串, pos=位置)
            if 匹配 is None:
                break
            位置 = 匹配.span()[1]
            表达式 = 匹配.group(1)
            if 表达式 is None:
                表达式 = 匹配.group(2)
            替代文本 = 语法分析器().分析('str(%s)' % 表达式)
            所有插值.append(替代文本.body[0].value)
            占位替代 = 占位替代.replace(匹配.group(0), '%s')

        if len(所有插值) == 0:
            return
        return 语法树.新节点(语法.二元表达式,
            左=语法树.新节点(语法.字符串, 值=占位替代, 片段=片段),
            运算符=ast.Mod(),
            右=语法树.多项(元素=所有插值, 上下文=ast.Load(), 片段=片段),
            片段=片段)

    @分析器母机.production(语法.常量.成分(名词_真))
    def 常量_真(self, 片段):
        return 语法树.常量(True, 片段=片段)

    @分析器母机.production(语法.常量.成分(名词_假))
    def 常量_假(self, 片段):
        return 语法树.常量(False, 片段=片段)

    @分析器母机.production(语法.常量.成分(名词_空))
    def 常量_空(self, 片段):
        return 语法树.常量(None, 片段=片段)

    @分析器母机.production(语法.常量.成分(名词_自身))
    def 常量_自身(self, 片段):
        return 语法树.新节点(语法.名称, 标识='self',
            上下文=ast.Load(),
            片段=片段)

    @分析器母机.production(语法.表达式.成分(语法.多项式乘法))
    @分析器母机.production(语法.表达式.成分(语法.二元表达式))
    @分析器母机.production(语法.表达式.成分(语法.一元表达式))
    @分析器母机.production(语法.表达式.成分(语法.表达式前缀))
    @分析器母机.production(语法.表达式.成分(语法.首要表达式))
    @分析器母机.production(语法.表达式.成分(语法.lambda表达式))
    @分析器母机.production(语法.表达式.成分(语法.三元表达式))
    @分析器母机.production(语法.表达式.成分(语法.数), precedence=等于)
    @分析器母机.production(语法.表达式.成分(语法.常量))

    # TODO 待参透：等于 的优先级在 连词_每隔 下一位
    @分析器母机.production(语法.表达式.成分(语法.范围表达式), precedence=等于)
    def 表达式(self, 片段):
        if 语法分析器.调试:
            print("表达式")
        return 片段[0]

    @分析器母机.production(语法.多项式乘法.成分(语法.数, 语法.表达式前缀))
    @分析器母机.production(语法.多项式乘法.成分(语法.数, 语法.首要表达式))
    def 多项式乘法(self, 片段):
        片段[1] = self.转换为多项(片段[1])
        return 语法树.新节点(语法.二元表达式,
            左=片段[0],
            运算符=ast.Mult(),
            右=片段[1],
            片段=片段)

    @分析器母机.production(语法.字典表达式.成分(前大括号, 冒号, 后大括号))
    @分析器母机.production(语法.字典表达式.成分(前大括号, 语法.各键值对, 后大括号))
    def 字典表达式(self, 片段):
        各键 = []
        各值 = []
        if isinstance(片段[1], list):
            各键 = [对[0] for 对 in 片段[1]]
            各值 = [对[1] for 对 in 片段[1]]
        return 语法树.字典(各键=各键,
            各值=各值,
            片段=片段)

    @分析器母机.production(语法.各键值对.成分(语法.键值对))
    @分析器母机.production(语法.各键值对.成分(语法.各键值对, 逗号, 语法.键值对))
    def 各键值对(self, 片段):
        if len(片段) < 3:
            return [片段[0]]
        片段[0].append(片段[-1])
        return 片段[0]

    @分析器母机.production(语法.键值对.成分(语法.表达式, 冒号, 语法.表达式))
    def 键值对(self, 片段):
        return (片段[-3], 片段[-1])

    @分析器母机.production(语法.列表表达式.成分(前中括号, 后中括号))
    @分析器母机.production(语法.列表表达式.成分(前中括号, 语法.各表达式, 后中括号))
    def 列表表达式(self, 片段):
        元素 = 片段[1] if len(片段) == 3 else []
        return 语法树.列表(元素, 片段=片段)

    @分析器母机.production(语法.各名称.成分(语法.名称))
    @分析器母机.production(语法.各名称.成分(语法.各名称, 逗号, 语法.名称))
    @分析器母机.production(语法.各实参.成分(语法.实参))
    @分析器母机.production(语法.各实参.成分(语法.各实参, 逗号, 语法.实参))
    @分析器母机.production(语法.各表达式.成分(语法.表达式))
    @分析器母机.production(语法.各表达式.成分(语法.各表达式, 逗号, 语法.表达式))
    def 各实参(self, 片段):
        if 语法分析器.调试:
            print("各实参")
        if len(片段) == 3:
            片段[0].append(片段[2])
            return 片段[0]
        return [片段[0]]

    @分析器母机.production(语法.实参.成分(语法.表达式))
    @分析器母机.production(语法.实参.成分(标识符, 符号_赋值, 语法.表达式))
    def 实参(self, 片段):
        if 语法分析器.调试:
            print("实参")
        if len(片段) == 1:
            return (片段[0], None)
        return (片段[-1], 片段[0].getstr())

    @分析器母机.production(语法.形参列表.成分())
    @分析器母机.production(语法.形参列表.成分(语法.非空形参列表))
    def 形参列表(self, 片段=[]):
        if not 片段:
            return 语法树.新节点(语法.形参列表, 参数=[])
        return self.形参合法化(片段[0])

    @分析器母机.production(语法.非空形参列表.成分(语法.形参))
    @分析器母机.production(语法.非空形参列表.成分(语法.非空形参列表, 逗号, 语法.形参))
    def 非空形参列表(self, 片段):
        if len(片段) == 1:
            各形参 = self.形参列表()
        else:
            各形参 = 片段[0]
        各形参.args.append(片段[-1])
        return 各形参

    @分析器母机.production(语法.形参.成分(语法.名称, 冒号, 语法.类型名称))
    @分析器母机.production(语法.形参.成分(语法.名称))
    def 形参(self, 片段):
        return 语法树.新节点(语法.形参,
            参数=片段[0].id,
            标注=(None if len(片段) == 1 else 片段[-1]),
            片段=片段)

    @分析器母机.production(语法.形参.成分(语法.名称, 符号_赋值, 语法.表达式))
    def 带默认值形参(self, 片段):
        参数 = self.形参(片段[:1])
        参数.default = 片段[2]
        return 参数

    @分析器母机.production(语法.函数.成分(名词_函数, 标识符, 前小括号, 语法.形参列表, 后小括号, 语法.块))
    @分析器母机.production(语法.函数.成分(名词_函数, 标识符, 语法.块))
    def 函数(self, 片段):
        return 语法树.新节点(语法.函数,
            名称=(片段[1].getstr()),
            参数=(片段[3] if len(片段) == 6 else self.形参列表()),
            主体=片段[-1],
            片段=片段)

    @分析器母机.production(语法.函数.成分(名词_函数, 标识符, 前小括号, 语法.形参列表, 后小括号, 冒号, 语法.类型名称, 语法.块))
    @分析器母机.production(语法.函数.成分(名词_函数, 标识符, 冒号, 语法.类型名称, 语法.块))
    def 函数(self, 片段):
        return 语法树.新节点(语法.函数,
            名称=(片段[1].getstr()),
            参数=(片段[3] if len(片段) == 8 else self.形参列表()),
            主体=片段[-1],
            返回=片段[-2],
            片段=片段)

    @分析器母机.production(语法.条件声明.成分(连词_如果, 语法.表达式, 语法.块, 语法.否则如果声明))
    @分析器母机.production(语法.条件声明.成分(连词_如果, 语法.表达式, 语法.块, 连词_否则, 语法.块))
    @分析器母机.production(语法.否则如果声明.成分())
    @分析器母机.production(语法.否则如果声明.成分(连词_否则如果, 语法.表达式, 语法.块, 语法.否则如果声明))
    @分析器母机.production(语法.否则如果声明.成分(连词_否则如果, 语法.表达式, 语法.块, 连词_否则, 语法.块))
    def 条件声明(self, 片段):
        if len(片段) == 0:
            return []
        否则部分 = 片段[-1]
        return 语法树.新节点(语法.条件声明,
            条件=片段[1],
            主体=片段[2],
            否则=否则部分 if isinstance(否则部分, list) else [否则部分],
            片段=片段)

    @分析器母机.production(语法.条件声明.成分(语法.声明, 连词_如果, 语法.表达式))
    def 条件倒置声明(self, 片段):
        return 语法树.新节点(语法.条件声明,
            条件=片段[-1],
            主体=[片段[0]],
            否则=[],
            片段=片段)

    @分析器母机.production(语法.声明.成分(语法.块))
    def 单块(self, 片段):
        return 语法树.新节点(语法.条件声明,
            条件=语法树.常量(True, 片段),
            主体=片段[0],
            否则=[],
            片段=片段)

    @分析器母机.production(语法.每当声明.成分(连词_每当, 语法.表达式, 语法.块))
    def 每当(self, 片段):
        return 语法树.新节点(语法.每当声明,
            条件=片段[1],
            主体=片段[2],
            片段=片段)

    @分析器母机.production(语法.每当声明.成分(动词_循环, 语法.块))
    def 无条件每当(self, 片段):
        return 语法树.新节点(语法.每当声明,
            条件=语法树.常量(True, 片段),
            主体=片段[1],
            片段=片段)

    @分析器母机.production(语法.迭代器.成分(语法.表达式前缀))
    @分析器母机.production(语法.迭代器.成分(语法.各表达式前缀))
    def 迭代器(self, 片段):
        return 片段[0]

    @分析器母机.production(语法.遍历范围.成分(语法.表达式))
    def 遍历范围(self, 片段):
        # TODO: 处理 ast.Starred
        return 片段[0]

    @分析器母机.production(语法.对于声明.成分(连词_对, 语法.迭代器, 连词_于, 语法.遍历范围, 语法.块))
    @分析器母机.production(语法.对于声明.成分(连词_对, 语法.迭代器, 冒号, 语法.遍历范围, 语法.块))
    def 对于声明(self, 片段):
        目标 = 片段[1]
        if isinstance(目标, list):
            for 项 in 目标:
                if hasattr(项, 'ctx'):
                    项.ctx = ast.Store()

            目标 = 语法树.多项(元素=目标,
               上下文=ast.Store(),
               片段=片段)
        else:
            目标.ctx = ast.Store()
        return 语法树.对于(目标=目标,
                      遍历范围=片段[3],
                      主体=片段[4],
                      片段=片段)

    @分析器母机.production(语法.对于声明.成分(语法.声明, 连词_对, 语法.迭代器, 连词_于, 语法.遍历范围))
    @分析器母机.production(语法.对于声明.成分(语法.声明, 连词_对, 语法.迭代器, 冒号, 语法.遍历范围))
    def 单个对于声明(self, 片段):
        倒装 = 片段[1:]
        倒装.append([片段[0]])
        return self.对于声明(倒装)

    @分析器母机.production(语法.名称.成分(标识符))
    def 标识符(self, 片段):
        标识 = 片段[0].getstr()
        名称 = 语法树.新节点(语法.名称,
            标识=标识,
            上下文=(ast.Load()),
            片段=片段)
        if not 标识.startswith('$'):
            return 名称
        名称.id = 'self'
        return 语法树.属性(
            值=名称,
            属性=标识.replace('$', ''),
            片段=片段)

    @分析器母机.error
    def error_handler(self, 词):
        if 词.getstr() == '\n':
            return
        # TODO: 最好取到语法信息(上下文)
        raise 语法错误(
            信息=('没认出这个词 "%s"' % 词.gettokentype()),
            文件名=self.文件名,
            行号=语法树.取行号(词),
            列号=语法树.取列号(词),
            源码=self.源码)

    def 形参合法化(self, 形参):
        带默认值 = False
        for 参数 in 形参.args:
            if hasattr(参数, 'default') and 参数.default:
                带默认值 = True
                形参.defaults.append(参数.default)
            elif 带默认值:
                raise 语法错误(信息='需要一个表达式指定默认值',
                  文件名=self.文件名,
                  行号=arg.lineno,
                  列号=arg.col_offset,
                  源码=self.源码)

        return 形参

    def 转换为多项(self, 各实参):
        if not isinstance(各实参, ast.arguments):
            return 各实参
        return 语法树.多项(上下文=ast.Load(),
            元素=[ast.Name((arg.arg), (ast.Load()), lineno=(arg.lineno), col_offset=(arg.col_offset)) for arg in 各实参.args],
            片段=各实参)

    分析器 = LRParser(分析器母机.build())

    def __init__(self, 分词器=分词器):
        self.分词器 = 分词器
        self.文件名 = ''
        self.匿名函数 = {}
        self.源码 = None

    def 分析(self, 源码, 源码文件=''):
        self.源码 = 源码.split("\n")
        语法树.源码 = self.源码
        self.文件名 = 源码文件
        try:
            各词 = self.分词器.lex(源码)
            # self.查看(各词)
            节点 = self.分析器.parse(各词, state=self)
        except LexingError as e:
            raise 词法错误(异常=e,
                    文件名=源码文件,
                    源码=源码)

        节点 = AnnoFuncInsertPass(self.匿名函数).visit(节点)
        节点 = NameFixPass(源码文件).visit(节点)
        return 节点

    def 查看(self, 各词):
        for 词 in 各词:
            print(f"{词}={词.getsourcepos()}")



import ast, sys
from io import StringIO
from pathlib import Path

from 木兰.分析器.词法分析器 import 分词器
from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.环境 import 创建全局变量
from 木兰.交互 import 开始交互
from 木兰.功用.反馈信息 import 中文化
from 木兰.功用.调试辅助 import 语法树相关

# 木兰不支持 compile, exec, 因此不改写为木兰代码
def 运行木兰代码(源码文件):
    with open(源码文件, 'r', encoding='utf-8') as f:
        源码 = f.read()

    分析器 = 语法分析器(分词器)
    节点 = 分析器.分析(源码, 源码文件)

    #print(ast.dump(节点, True, True))
    #print(语法树相关.格式化节点(节点, 1))

    # 参考： https://stackoverflow.com/questions/3906232/python-get-the-print-output-in-an-exec-statement
    原标准输出 = sys.stdout
    重定向输出 = sys.stdout = StringIO()

    # 参考：https://docs.python.org/3.7/library/functions.html?highlight=compile#compile
    try:
        可执行码 = compile(节点, 源码文件, 'exec')

        环境变量 = 创建全局变量(文件名=源码文件)

        try:
            exec(可执行码, 环境变量)
        except Exception as e:
            # TODO: 提神符为确保各平台显示一致, 改为图片
            return " 😰 " + "\n".join([行 for 行 in 中文化(e, 源码文件) if 行.find(str(Path("编辑器/运行木兰.py"))) < 0])
    except SyntaxError as 语法错误:
        return f"语法错误: {语法错误}\n"
    finally:
        sys.stdout = 原标准输出

    return 重定向输出.getvalue()

#print(运行木兰代码("测试/数据结构/字典.ul"))


import math, sys, imp, os, threading
from pathlib import Path

from 木兰.分析器.语法分析器 import 语法分析器


def 分析并编译(源码文件名):
    with open(源码文件名, 'r', encoding='utf-8') as 文件:
        源码 = 文件.read()

        分析器 = 语法分析器()
        节点 = 分析器.分析(源码=源码, 源码文件=源码文件名)
        return compile(节点, 源码文件名, 'exec')

def 加载木兰模块(名称, 全局, 源自=(), 目录相对层次=0):
    木兰源码路径 = str(Path(*(名称.split(".")))) + '.ul'
    可执行码 = 分析并编译(木兰源码路径)
    if 可执行码 is None:
        raise ModuleNotFoundError(名称)

    # TODO: 研究何用
    所有模块 = []
    后段 = 名称
    模块名 = lambda 名称: 所有模块[-1].__name__ + '.' + 名称 if 所有模块 else ""

    点位 = 0
    while 点位 != -1:
        点位 = 后段.find('.')
        前段, 后段 = 后段[:点位], 后段[点位 + 1:]
        if 点位 == -1:
            前段 = 后段
            模块 = imp.new_module(模块名(后段))
            模块.__dict__.update(创建全局变量(argv=(全局['ARGV'])))
            模块.__dict__['__file__'] = os.path.abspath(木兰源码路径)
            exec(可执行码, 模块.__dict__)
        else:
            模块 = imp.new_module(模块名(前段))
        if 所有模块:
            所有模块[-1].__dict__[前段] = 模块
        所有模块.append(模块)

    顶层 = 所有模块[0]
    if len(所有模块) > 1:
        if 源自 is not None:
            for sym in 源自:
                if sym == '*':
                    for k in 模块.__dict__:
                        if k not in 全局:
                            顶层.__dict__[k] = 模块.__dict__[k]
                    break
                else:
                    顶层.__dict__[sym] = 模块.__dict__[sym]

    return 顶层

def 内置扩展(内置项):
    from inspect import isclass
    for k, v in __builtins__.items():
        if isclass(v) and issubclass(v, BaseException):
            内置项[k] = v

    return 内置项

def __内置_除(a, b):
    if isinstance(a, int):
        if isinstance(b, int):
            return math.floor(a / b)
    return a / b

def __内置_求余(a, b):
    # TODO: 不解为何要在 a b 为整数时特殊处理
    return a % b

def 创建全局变量(argv=[], 文件名=''):

    def 转字符串(x):

        def 容器转为字符串(容器, 始='', 末=''):
            字符串 = 始
            for 序号, 该项 in enumerate(容器):
                if 序号:
                    字符串 += ', '
                字符串 += 转字符串(该项)

            字符串 += 末
            return 字符串

        if x is None:
            return 'nil'
        if isinstance(x, bool):
            return 'true' if x else 'false'
        if isinstance(x, list):
            return 容器转为字符串(x, '[', ']')
        if isinstance(x, tuple):
            return 容器转为字符串(x)
        return str(x)

    def 自定义输出(*各物件, 分隔符=' ', 终止符='', 文件=sys.stdout, flush=False):
        """打印内容到流, 默认到标准输出."""
        for 物件 in 各物件:
            文件.write(转字符串(物件))
            if 物件 != 各物件[(-1)]:
                文件.write(分隔符)

        文件.write(终止符)
        if flush:
            文件.flush()

    def 自定义导入(名称, 全局=None, 本地=None, 源自=(), 目录相对层次=0):
        """导入一个木兰模块, 如果没找到, 则导入 Python 模块
        """
        try:
            return 加载木兰模块(名称, 全局, 源自, 目录相对层次)
        except FileNotFoundError:
            return __import__(名称, 全局, 本地, 源自, 目录相对层次)
        except Exception as 例外:
            raise 例外

    def 本地断言(表达式, 反馈=None):
        assert 表达式, 反馈

    def 内置自身():
        """ 当前任务 ID """
        return threading.currentThread()

    def eval_print(expr):
        if expr is None:
            return
        try:
            expr()
        except Exception:
            自定义输出(expr, 终止符='\n')

    当前目录 = os.getcwd()
    if 当前目录 not in sys.path:
        sys.path.append(当前目录)
    return {
        'print': 自定义输出,
        'println': lambda *各物件: 自定义输出(*各物件, **{'终止符': '\n'}),
        'assert': 本地断言,
        'enumerate':enumerate,
        'len': len,
        'all': all,
        'any': any,
        'range': range,
        'input':input,
        'super':super,
        'int': int,
        'str': str,
        'list': list,
        'set': set,
        'tuple':lambda *各实参: 各实参,
        'isa': lambda x, 类型: isinstance(x, 类型),
        'max': max,
        'min': min,
        'map': map,
        'filter': filter,
        'zip':zip,
        'staticmethod': staticmethod,
        'property':property,
        'self':内置自身,
        '再会':sys.exit,
        'quit':sys.exit,
        'open':open,
        'ARGV': argv,
        '__builtins__': 内置扩展({
            '__import__': 自定义导入,
            '__build_class__': __build_class__,
            '__name__': '__main__',
            '__file__': 文件名,
            '__print__': eval_print,
            '__div__': __内置_除,
            '__rem__': __内置_求余})
    }


from 测试.期望值表 import 期望值
from 编辑器.运行木兰 import 运行木兰代码

import unittest

class test所有(unittest.TestCase):

    def test(self):
        测试目录 = '测试/'
        全部通过 = True
        失败表 = {}

        for 文件 in 期望值:
            路径 = 测试目录 + 文件
            实际值 = 运行木兰代码(路径)
            预期值 = 期望值[文件].decode("utf-8")

            if 实际值 != 预期值:
                失败表[文件] = 实际值
                全部通过 = False

        for 文件 in 失败表:
            print(f"失败： {文件} 期望：{repr(期望值[文件].decode('utf-8'))} 实际：{repr(失败表[文件])}")
        self.assertTrue(全部通过, "以上用例未通过！")



import setuptools

with open("README.md", encoding="utf-8", mode="r") as readme:
    longDescription = readme.read()

setuptools.setup(
    name="ulang",
    # 木兰/__init__.py 中保持一致
    version="0.0.16",
    license="GNU GPLv3",
    author="吴烜",
    author_email="mulanrevive@gmail.com",
    entry_points = {
        "console_scripts": ['木兰 = 木兰.中:中']
        },
    description="木兰编程语言重现",
    long_description=longDescription,
    long_description_content_type="text/markdown",
    url="https://www.oschina.net/p/mulan-rework",
    project_urls={
        "Source": "https://gitee.com/MulanRevive/mulan-rework",
        "Changelog": "https://gitee.com/MulanRevive/mulan-rework#版本历史介绍",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'rply',
    ],
    python_requires="== 3.7.*"
)


import ast

from rply.token import SourcePosition
from rply import Token

from 木兰.分析器.语法成分 import *

class 语法树:

    源码 = []

    @staticmethod
    def 新节点(类型, 主体=None, 忽略类型=None, 值=None, 左=None, 运算符=None, 右=None, 标识=None,
            上下文=None, 函数=None, 参数=None, 关键词=None, 变量=None, 条件=None, 否则=None,
            前项=None, 后项=None, 标注=None, 名称=None, 返回=None, 各基准类=None, 片段=None):
        if 类型 == 语法.模块:
            节点 = ast.Module(body=主体, type_ignores=忽略类型)
        elif 类型 == 语法.表达式:
            节点 = ast.Expr(value=值)
        elif 类型 == 语法.数:
            节点 = ast.Num(n=值)
        elif 类型 == 语法.二元表达式:
            if isinstance(运算符, ast.And) or isinstance(运算符, ast.Or):
                节点 = ast.BoolOp(op=运算符, values=[前项, 后项])
            elif isinstance(运算符, ast.Add) or isinstance(运算符, ast.Sub) or isinstance(运算符, ast.Mult) or isinstance(运算符, ast.Pow) or isinstance(运算符, ast.Mod):
                节点 = ast.BinOp(left=左, op=运算符, right=右)
            else:
                # TODO: 为何比较符和后项在数组中?
                节点 = ast.Compare(前项, [运算符], [后项])
        elif 类型 == 语法.名称:
            节点 = ast.Name(id=标识, ctx=上下文)
        elif 类型 == 语法.调用:
            节点 = ast.Call(func=函数, args=参数, keywords=关键词)
        elif 类型 == 语法.赋值:
            节点 = ast.Assign([变量], 值)
        elif 类型 == 语法.增量赋值:
            节点 = ast.AugAssign(变量, 运算符, 值)
        elif 类型 == 语法.条件声明:
            节点 = ast.If(test=条件, body=主体, orelse=否则)
        elif 类型 == 语法.每当声明:
            节点 = ast.While(test=条件, body=主体, orelse=[])
        elif 类型 == 语法.终止声明:
            节点 = ast.Break()
        elif 类型 == 语法.跳过声明:
            节点 = ast.Continue()
        elif 类型 == 语法.操作数:
            节点 = ast.arg(arg=参数)
        elif 类型 == 语法.lambda形参 or 类型 == 语法.形参:
            节点 = ast.arg(arg=参数, annotation=标注)
        elif 类型 == 语法.形参列表:
            节点 = ast.arguments(args=参数, kwonlyargs=[], kw_defaults=[], defaults=[], vararg=None, kwarg=None)
        elif 类型 == 语法.函数:
            节点 = ast.FunctionDef(
                    name=名称,
                    args=参数,
                    body=主体,
                    decorator_list=[])
            if 返回:
                节点.returns = 返回
        elif 类型 == 语法.返回声明:
            节点 = ast.Return(value=值)
        elif 类型 == 语法.引用声明:
            节点 = ast.Import(names=名称)
        elif 类型 == 语法.lambda表达式:
            节点 = ast.Lambda(args=参数, body=主体)
        elif 类型 == 语法.类型定义:
            节点 = ast.ClassDef(name=名称,
                            bases=各基准类,
                            keywords=[],
                            body=主体,
                            decorator_list=[])
        elif 类型 == 语法.字符串:
            节点 = ast.Str(值)
        elif 类型 == 语法.外部声明:
            节点 = ast.Global(names=名称)
        elif 类型 == 语法.一元表达式:
            节点 = ast.UnaryOp(op=运算符, operand=值)

        if 片段:
            节点.lineno = 语法树.取行号(片段)
            节点.col_offset = 语法树.取列号(片段)
        return 节点

    @staticmethod
    def 别名(名称, 别名, 片段):
        return ast.alias(name=名称, asname=别名, lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 属性(值, 属性, 片段):
        return ast.Attribute(value=值, attr=属性, ctx=(ast.Load()), lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 从模块导入(模块, 各名称, 位置, 片段):
        return ast.ImportFrom(module=模块, names=各名称, level=位置, lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 常量(值, 片段):
        return ast.NameConstant(value=值, lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 空转(片段):
        return ast.Pass(lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 对于(目标, 遍历范围, 主体, 片段):
        return ast.For(target=目标, iter=遍历范围, body=主体, orelse=[],
                       lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 列表(元素, 片段):
        return ast.List(elts=元素, ctx=ast.Load(), lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 索引(值, 片段):
        return ast.Index(value=值, lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    @staticmethod
    def 片(下限, 上限, 片段):
        return ast.Slice(lower=下限,
                        upper=上限,
                        step=None,
                        lineno=语法树.取行号(片段),
                        col_offset=语法树.取列号(片段))
    @staticmethod
    def 下标(全值, 片, 片段):
        return ast.Subscript(value=全值,
                             slice=片,
                             ctx=(ast.Load()),
                             lineno=语法树.取行号(片段),
                             col_offset=语法树.取列号(片段))

    @staticmethod
    def 多项(元素, 上下文, 片段):
        return ast.Tuple(elts=元素,
                         ctx=上下文,
                         lineno=语法树.取行号(片段),
                         col_offset=语法树.取列号(片段))

    @staticmethod
    def 字典(各键, 各值, 片段):
        return ast.Dict(keys=各键,
            values=各值,
            lineno=语法树.取行号(片段),
            col_offset=语法树.取列号(片段))

    @staticmethod
    def 顺便处理项(上下文表达式, 可选变量, 片段):
        return ast.withitem(context_expr=上下文表达式,
            optional_vars=可选变量,
            lineno=语法树.取行号(片段),
            col_offset=语法树.取列号(片段))

    @staticmethod
    def 顺便(各项, 主体, 片段):
        return ast.With(items=各项,
            body=主体,
            lineno=语法树.取行号(片段),
            col_offset=语法树.取列号(片段))

    @staticmethod
    def 试试(主体, 处理, 片段):
        return ast.Try(body=主体,
                       handlers=处理,
                       orelse=[],
                       finalbody=[],
                       lineno=语法树.取行号(片段),
                       col_offset=语法树.取列号(片段))

    @staticmethod
    def 例外处理(类型, 名称, 主体, 片段):
        return ast.ExceptHandler(type=类型,
                                 name=名称,
                                 body=主体,
                                 lineno=语法树.取行号(片段),
                                 col_offset=语法树.取列号(片段))
    @staticmethod
    def 引发(例外, 片段):
        return ast.Raise(exc=例外,
            cause=None,
            lineno=语法树.取行号(片段),
            col_offset=语法树.取列号(片段))

    # 表达式部分

    @staticmethod
    def 如果表达式(条件, 主体, 否则, 片段):
        return ast.IfExp(test=条件, body=主体, orelse=否则, lineno=语法树.取行号(片段), col_offset=语法树.取列号(片段))

    '''
    不同于 python3 的语法树中, col_offset 是从 0 开始:
    >>> ast.dump(ast.parse("2+3"), True, True)
    'Module(body=[Expr(value=BinOp(left=Num(n=2, lineno=1, col_offset=0), op=Add(), right=Num(n=3, lineno=1, col_offset=2), lineno=1, col_offset=0), lineno=1, col_offset=0)])'
    '''

    def 取源码位置(片段):
        if isinstance(片段, list):
            if len(片段) > 0:
                片段 = 片段[0]
        if isinstance(片段, Token):
            if 片段.gettokentype() == '$end':
                return 语法树.取末位()
            return 片段.getsourcepos()
        # Constant 也是 ast.expr
        if isinstance(片段, ast.stmt) or isinstance(片段, ast.expr):
            # TODO: 之前没 import SourcePosition 时, 编译/运行未报错! 需解决
            return SourcePosition(0, 片段.lineno, 片段.col_offset)
        return SourcePosition(0, 0, 0)

    def 取末位():
        idx = -1
        行号 = len(语法树.源码)
        列号 = len(语法树.源码[(-1)])
        return SourcePosition(idx, 行号, 列号)

    def 取行号(片段):
        try:
            return 语法树.取源码位置(片段).lineno
        except:
            return 0

    def 取列号(片段):
        try:
            return 语法树.取源码位置(片段).colno
        except:
            return 0



#!/usr/bin/env python3

import os, sys, getopt
import ast
from 木兰.分析器.词法分析器 import 分词器
from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.环境 import 创建全局变量
from 木兰.交互 import 开始交互
from 木兰.功用.反馈信息 import 反馈信息
from 木兰.功用.调试辅助 import 语法树相关

def 用途(程序):
    介绍 = '''使用方法: %s 源码文件
选项:
 --版本,         -版   显示版本
'''
    sys.stderr.write(介绍 % os.path.basename(程序))
    sys.exit(-1)

def 中(argv=None):
    if argv is None:
        argv = sys.argv

    try:
        选项, 参数 = getopt.getopt(argv[1:], '版', [
            '版本'])
    except getopt.GetoptError as e:
        try:
            sys.stderr.write(str(e) + '\n')
            用途(argv[0])
        finally:
            e = None
            del e

    版本 = False
    for 某项, 值 in 选项:
        if 某项 in ('-版', '--版本'):
            版本 = True

    if 版本:
        from 木兰 import __版本__
        sys.stderr.write('%s\n' % __版本__)
        sys.exit()

    if len(sys.argv) == 1:
        sys.exit(开始交互())

    if len(参数) > 0:
        源码文件 = 参数[0]
    with open(源码文件, 'r', encoding='utf-8') as f:
        源码 = f.read()

    分析器 = 语法分析器(分词器)
    节点 = 分析器.分析(源码, 源码文件)

    #print(ast.dump(节点, True, True))
    #print(语法树相关.格式化节点(节点, 1))

    # 参考：https://docs.python.org/3.7/library/functions.html?highlight=compile#compile
    try:
        可执行码 = compile(节点, 源码文件, 'exec')

        环境变量 = 创建全局变量(文件名=源码文件)

        try:
            exec(可执行码, 环境变量)
        except Exception as e:
            try:
                sys.stderr.write('%s\n' % 反馈信息(e, 源码文件))
            finally:
                e = None
                del e
    except SyntaxError as 语法错误:
        sys.stderr.write(f"语法错误: {语法错误}\n")



from rply import LexerGenerator
import rply, re

# 顺序与词法规则添加顺序一致. 值不为中文的是关键词.

小数 = '小数'
整数 = '整数'
双引号字符串 = '双引号字符串'
单引号字符串 = '单引号字符串'
前大括号 = '前大括号'
后大括号 = '后大括号'
名词_空 = 'nil'
名词_真 = 'true'
名词_假 = 'false'
连词_且 = 'and'
连词_或 = 'or'
连词_如果 = 'if'
连词_否则如果 = 'elif'
连词_否则 = 'else'
连词_每当 = 'while'
动词_循环 = 'loop'
连词_对 = 'for'
动词_返回 = 'return'
动词_终止 = 'break'
动词_跳过 = 'continue'
名词_函数 = 'func'
名词_类型 = 'type'
动词_引用 = 'using'
连词_于 = 'in'
动词_试试 = 'try'
动词_接手 = 'catch'
动词_抛出 = 'throw'
名词_操作符 = 'operator'
连词_每隔 = 'by'
形容词_外部 = 'extern'
名词_超类 = 'super'
名词_应变属性 = 'attr'
标识符 = '标识符'
点点小于 = '..<'
点点 = '..'
点 = '.'
名词_自身 = '$'
前中括号 = '['
后中括号 = ']'
前小括号 = '('
后小括号 = ')'
严格等于 = '==='
严格不等于 = '!=='
等于 = '=='
不等于 = '!='
大于等于 = '>='
小于等于 = '<='
箭头 = '->'
大于 = '>'
小于 = '<'
符号_赋值 = '='
逗号 = ','
增量赋值 = '+='
减量赋值 = '-='
加 = '+'
减 = '-'
星号 = '*'
除 = '/'
求余 = '%'
乘方 = '^'
分号 = ';'
非 = '!'
问号 = '?'
冒号 = ':'
取反 = '~'
换行 = '换行'

规则 = [
    小数,
    整数,
    加,
    减,
    星号,
    除,
    标识符,
    前小括号,
    后小括号,
    严格等于,
    严格不等于,
    等于,
    不等于,
    大于,
    小于,
    大于等于,
    小于等于,
    换行,
    符号_赋值,
    逗号,
    前大括号,
    后大括号,
    连词_如果,
    连词_否则如果,
    连词_否则,
    连词_或,
    连词_且,
    连词_每当,
    动词_终止,
    动词_跳过,
    名词_函数,
    动词_返回,
    问号,
    冒号,
    动词_引用,
    点,
    连词_于,
    名词_空,
    非,
    名词_类型,
    名词_操作符,
    形容词_外部,
    名词_真,
    名词_假,
    双引号字符串,
    单引号字符串,
    连词_对,
    点点,
    点点小于,
    分号,
    增量赋值,
    减量赋值,
    动词_循环,
    前中括号,
    后中括号,
    连词_每隔,
    箭头,
    乘方,
    名词_自身,
    名词_应变属性,
    动词_试试,
    求余,
    名词_超类,
    取反,
    动词_抛出,
    动词_接手,
]

分词器母机 = LexerGenerator()

分词器母机.add(小数, r'\d+\.\d+')
分词器母机.add(整数, r'\d+')
分词器母机.add(双引号字符串, r'(")((?<!\\)\\\1|.)*?\1')
分词器母机.add(单引号字符串, r"(')((?<!\\)\\\1|.)*?\1") # \' 之前不能有 \
分词器母机.add(前大括号, r'{\r*\n*') # TODO: 何用？ , flags=(re.DOTALL)
# TODO: 参考[rply 测试用例](https://github.com/alex/rply/blob/19a9e08c486b2723a2e2378df6edb6a26e2df4a5/tests/test_lexer.py#L106)
# 不知在正则表达式中无`.`时此 flag 有何用
分词器母机.add(后大括号, r'\r*\n*}') # , flags=(re.DOTALL)
分词器母机.add(名词_空, r'\bnil\b')
分词器母机.add(名词_真, r'\btrue\b')
分词器母机.add(名词_假, r'\bfalse\b')
分词器母机.add(连词_且, r'\band\b')
分词器母机.add(连词_或, r'\bor\b')
分词器母机.add(连词_如果, r'\bif\b')
分词器母机.add(连词_否则如果, r'\r*\n*\s*elif\s*\r*\n*') # TODO: 何用？ , flags=(re.DOTALL)
分词器母机.add(连词_否则, r'\r*\n*\s*else\s*\r*\n*') # , flags=(re.DOTALL)
分词器母机.add(连词_每当, r'\bwhile\b')
分词器母机.add(动词_循环, r'\bloop\b')
分词器母机.add(连词_对, r'\bfor\b')
分词器母机.add(动词_返回, r'\breturn\b')
分词器母机.add(动词_终止, r'\bbreak\b')
分词器母机.add(动词_跳过, r'\bcontinue\b')
分词器母机.add(名词_函数, r'\bfunc\b')
分词器母机.add(名词_类型, r'\btype\b')
分词器母机.add(动词_引用, r'\busing\b')
分词器母机.add(连词_于, r'\bin\b')
分词器母机.add(动词_试试, r'\btry\b')
分词器母机.add(动词_接手, r'\r*\n*\s*catch\s*\r*\n*') # , flags=(re.DOTALL)
分词器母机.add(动词_抛出, r'\bthrow\b')
分词器母机.add(名词_操作符, r'\boperator\b')
分词器母机.add(连词_每隔, r'\bby\b')
分词器母机.add(形容词_外部, r'\bextern\b')
分词器母机.add(名词_超类, r'\bsuper\b')
分词器母机.add(名词_应变属性, r'\battr\b')
分词器母机.add(标识符, r'\$?[_a-zA-Z\u4e00-\u9fa5][_a-zA-Z0-9\u4e00-\u9fa5]*')
分词器母机.add(点点小于, r'\.\.<')
分词器母机.add(点点, r'\.\.')
分词器母机.add(点, r'\.')
分词器母机.add(名词_自身, r'\$')
分词器母机.add(前中括号, r'\[')
分词器母机.add(后中括号, r'\]')
分词器母机.add(前小括号, r'\(')
分词器母机.add(后小括号, r'\)')
分词器母机.add(严格等于, '===')
分词器母机.add(严格不等于, '!==')
分词器母机.add(等于, '==')
分词器母机.add(不等于, '!=')
分词器母机.add(大于等于, '>=')
分词器母机.add(小于等于, '<=')
分词器母机.add(箭头, '->')
分词器母机.add(大于, '>')
分词器母机.add(小于, '<')
分词器母机.add(符号_赋值, '=')
分词器母机.add(逗号, ',')
分词器母机.add(增量赋值, r'\+=')
分词器母机.add(减量赋值, '-=')
分词器母机.add(加, '\+')
分词器母机.add(减, '-')
分词器母机.add(星号, '\*')
分词器母机.add(除, '/')
分词器母机.add(求余, '%')
分词器母机.add(乘方, r'\^')
分词器母机.add(分号, ';')
分词器母机.add(非, '!')
分词器母机.add(问号, r'\?')
分词器母机.add(冒号, ':')
分词器母机.add(取反, '~')
分词器母机.add(换行, r'\n')
分词器母机.ignore('[ \t]+') # TODO: \r 何用? 也许和 windows 换行有关
分词器母机.ignore('//[^\n]*')
分词器母机.ignore('/\\*.*?\\*/', flags=(re.DOTALL))

分词器 = 分词器母机.build()
